# coding: utf-8
import asyncio
import sys
import warnings

from .utils import run_coroutine

import asynctest.mock


# add a new-style coroutine to the Test class:
def patch_Test_Class(klass):
    class Test(klass):
        async def an_async_coroutine(self):
            pass

    return Test


# wrap the coro in a new-style coroutine
def transform(coro):
    async def a_coroutine(*a, **kw):
        return await coro(*a, **kw)

    return a_coroutine


# make a simple coroutine which invokes a function before awaiting on an
# awaitable and after
def build_simple_coroutine(before_func, after_func=None):
    async def a_coroutine(awaitable, *args, **kwargs):
        before = before_func(*args, **kwargs)
        await awaitable
        after = (after_func or before_func)(*args, **kwargs)
        return before, after

    return a_coroutine


class _Test_Mock_Of_Async_Magic_Methods:
    class WithAsyncContextManager:
        def __init__(self):
            self.entered = False
            self.exited = False

        async def __aenter__(self, *args, **kwargs):
            self.entered = True
            return self

        async def __aexit__(self, *args, **kwargs):
            self.exited = True

    def test_mock_magic_methods_are_coroutine_mocks(self, klass):
        for spec in (None, self.WithAsyncContextManager()):
            with self.subTest(spec=spec):
                mock_instance = asynctest.mock.MagicMock(spec)
                self.assertIsInstance(mock_instance.__aenter__,
                                      asynctest.mock.CoroutineMock)
                self.assertIsInstance(mock_instance.__aexit__,
                                      asynctest.mock.CoroutineMock)

    def test_mock_supports_async_context_manager(self, klass):
        called = False
        instance = self.WithAsyncContextManager()
        mock_instance = asynctest.mock.MagicMock(instance)

        async def use_context_manager():
            nonlocal called
            async with mock_instance as result:
                called = True

            return result

        result = run_coroutine(use_context_manager())
        self.assertFalse(instance.entered)
        self.assertFalse(instance.exited)
        self.assertTrue(called)
        self.assertTrue(mock_instance.__aenter__.called)
        self.assertTrue(mock_instance.__aexit__.called)
        self.assertIsNot(mock_instance, result)
        self.assertIsInstance(result, asynctest.mock.MagicMock)

    def test_mock_customize_async_context_manager(self, klass):
        instance = self.WithAsyncContextManager()
        mock_instance = asynctest.mock.MagicMock(instance)

        expected_result = object()
        mock_instance.__aenter__.return_value = expected_result

        async def use_context_manager():
            async with mock_instance as result:
                return result

        self.assertIs(run_coroutine(use_context_manager()), expected_result)

    def test_mock_customize_async_context_manager_with_coroutine(self, klass):
        enter_called = False
        exit_called = False

        async def enter_coroutine(*args):
            nonlocal enter_called
            enter_called = True

        async def exit_coroutine(*args):
            nonlocal exit_called
            exit_called = True

        instance = self.WithAsyncContextManager()
        mock_instance = asynctest.mock.MagicMock(instance)

        mock_instance.__aenter__ = enter_coroutine
        mock_instance.__aexit__ = exit_coroutine

        async def use_context_manager():
            async with mock_instance:
                pass

        run_coroutine(use_context_manager())
        self.assertTrue(enter_called)
        self.assertTrue(exit_called)

    def test_context_manager_raise_exception_by_default(self, klass):
        class InContextManagerException(Exception):
            pass

        async def raise_in(context_manager):
            async with context_manager:
                raise InContextManagerException()

        instance = self.WithAsyncContextManager()
        mock_instance = asynctest.mock.MagicMock(instance)
        with self.assertRaises(InContextManagerException):
            run_coroutine(raise_in(mock_instance))

    class WithAsyncIterator:
        def __init__(self):
            self.iter_called = False
            self.next_called = False
            self.items = ["foo", "bar", "baz"]

        def __aiter__(self):
            return self

        async def __anext__(self):
            try:
                return self.items.pop()
            except IndexError:
                pass

            raise StopAsyncIteration

    class WithAsyncIteratorDeprecated(WithAsyncIterator):
        # Before python 3.5.2, __aiter__ is specified as a coroutine, but it's
        # a design error, it should be a plain function.
        async def __aiter__(self):
            return super().__aiter__()

    def get_async_iterator_classes(self):
        # We assume that __aiter__ as a coroutine will not be available in
        # python 3.7, see: pep-0525#aiter-and-anext-builtins
        if sys.version_info >= (3, 7):
            return (self.WithAsyncIterator, )
        else:
            return (self.WithAsyncIterator, self.WithAsyncIteratorDeprecated, )

    def test_mock_aiter_and_anext(self, klass):
        classes = self.get_async_iterator_classes()

        for iterator_class in classes:
            with self.subTest(iterator_class=iterator_class.__name__):
                instance = iterator_class()
                mock_instance = asynctest.MagicMock(instance)

                self.assertEqual(asyncio.iscoroutine(instance.__aiter__),
                                 asyncio.iscoroutine(mock_instance.__aiter__))
                self.assertEqual(asyncio.iscoroutine(instance.__anext__),
                                 asyncio.iscoroutine(mock_instance.__anext__))

                iterator = instance.__aiter__()
                if asyncio.iscoroutine(iterator):
                    iterator = run_coroutine(iterator)

                mock_iterator = mock_instance.__aiter__()
                if asyncio.iscoroutine(mock_iterator):
                    mock_iterator = run_coroutine(mock_iterator)

                self.assertEqual(asyncio.iscoroutine(iterator.__aiter__),
                                 asyncio.iscoroutine(mock_iterator.__aiter__))
                self.assertEqual(asyncio.iscoroutine(iterator.__anext__),
                                 asyncio.iscoroutine(mock_iterator.__anext__))

    def test_mock_async_for(self, klass):
        async def iterate(iterator):
            accumulator = []
            # don't print the DeprecationWarning for __aiter__
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                async for item in iterator:
                    accumulator.append(item)

            return accumulator

        expected = ["FOO", "BAR", "BAZ"]
        specs = [None]
        specs.extend(i() for i in self.get_async_iterator_classes())
        for spec in specs:
            with self.subTest("iterate through default value"):
                mock_instance = asynctest.MagicMock(spec)
                self.assertEqual([], run_coroutine(iterate(mock_instance)))

            with self.subTest("iterate through set return_value"):
                mock_instance = asynctest.MagicMock(spec)
                mock_instance.__aiter__.return_value = expected[:]

                self.assertEqual(expected, run_coroutine(iterate(mock_instance)))

            with self.subTest("iterate through set return_value iterator"):
                mock_instance = asynctest.MagicMock(spec)
                mock_instance.__aiter__.return_value = iter(expected[:])

                self.assertEqual(expected, run_coroutine(iterate(mock_instance)))



# coding: utf-8
import asyncio


def run_coroutine(coroutine):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        return loop.run_until_complete(coroutine)
    finally:
        loop.close()



# coding: utf-8

import asynctest


# test_coroutinefunction_executed
class CoroutineFunctionTest(asynctest.TestCase):
    ran = False

    async def noop(self):
        pass

    async def runTest(self):
        self.ran = True
        await self.noop()


# test_coroutine_returned_executed
class CoroutineTest(CoroutineFunctionTest):
    ran = False

    def runTest(self):
        return super().runTest()



# coding: utf-8

import asyncio
import itertools
import logging
import os
import re
import subprocess
import sys
import time
import unittest
import unittest.mock

import asynctest

if sys.version_info >= (3, 5):
    from . import test_case_await as _using_await
else:
    _using_await = None


class Test:
    class FooTestCase(asynctest.TestCase):
        def runTest(self):
            pass

        def test_foo(self):
            pass

    @asynctest.fail_on(unused_loop=False)
    class LoggingTestCase(asynctest.TestCase):
        def __init__(self, calls):
            super().__init__()
            self.calls = calls

        def setUp(self):
            self.calls.append('setUp')

        def test_basic(self):
            self.events.append('test_basic')

        def tearDown(self):
            self.events.append('tearDown')

    class StartWaitProcessTestCase(asynctest.TestCase):
        @staticmethod
        @asyncio.coroutine
        def start_wait_process(loop):
            process = yield from asyncio.create_subprocess_shell(
                "true", stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                loop=loop)

            try:
                out, err = yield from asyncio.wait_for(
                    process.communicate(), timeout=.1, loop=loop)
            except BaseException:
                process.kill()
                os.waitpid(process.pid, os.WNOHANG)
                raise

        @asyncio.coroutine
        def runTest(self):
            yield from self.start_wait_process(self.loop)


class _TestCase(unittest.TestCase):
    run_methods = ('run', 'debug', )

    def create_default_loop(self):
        default_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(default_loop)
        self.addCleanup(default_loop.close)
        return default_loop


class Test_TestCase(_TestCase):
    def test_init_and_close_loop_for_test(self):
        default_loop = self.create_default_loop()

        @asynctest.lenient
        class LoopTest(asynctest.TestCase):
            failing = False

            def runTest(self):
                try:
                    self.assertIsNotNone(self.loop)
                    self.assertFalse(self.loop.close.called)
                except Exception:
                    self.failing = True
                    raise

            def runFailingTest(self):
                self.runTest()
                raise SystemError()

        for method, test in itertools.product(self.run_methods, ('runTest', 'runFailingTest', )):
            with self.subTest(method=method, test=test), \
                    unittest.mock.patch('asyncio.new_event_loop') as mock:
                mock_loop = unittest.mock.Mock(asyncio.AbstractEventLoop)
                mock.return_value = mock_loop

                case = LoopTest(test)
                try:
                    getattr(case, method)()
                except SystemError:
                    pass

                mock.assert_called_with()
                mock_loop.close.assert_called_with()

                if sys.version_info >= (3, 6):
                    mock_loop.run_until_complete.__wrapped__.assert_called_with(mock_loop.shutdown_asyncgens())

                self.assertFalse(case.failing)
                self.assertIs(default_loop, asyncio.get_event_loop())

    def test_default_loop_is_not_created_when_unused(self):
        policy = asyncio.get_event_loop_policy()

        @asynctest.fail_on(unused_loop=False)
        class Dummy_TestCase(Test.FooTestCase):
            pass

        for method in self.run_methods:
            with self.subTest(method=method):
                with unittest.mock.patch.object(policy,
                                                "get_event_loop") as mock:
                    case = Dummy_TestCase()
                    getattr(case, method)()

                self.assertFalse(mock.called)

    def test_update_default_loop_works(self):
        a_loop = asyncio.new_event_loop()
        self.addCleanup(a_loop.close)

        class Update_Default_Loop_TestCase(asynctest.TestCase):
            @asynctest.fail_on(unused_loop=False)
            def runTest(self):
                self.assertIs(self.loop, asyncio.get_event_loop())
                asyncio.set_event_loop(a_loop)
                self.assertIs(a_loop, asyncio.get_event_loop())

        for method in self.run_methods:
            with self.subTest(method=method):
                case = Update_Default_Loop_TestCase()
                result = getattr(case, method)()

                if result:
                    self.assertTrue(result.wasSuccessful())

                self.assertIs(a_loop, asyncio.get_event_loop())

    def test_use_default_loop(self):
        default_loop = self.create_default_loop()

        class Using_Default_Loop_TestCase(asynctest.TestCase):
            use_default_loop = True

            @asynctest.fail_on(unused_loop=False)
            def runTest(self):
                self.assertIs(default_loop, self.loop)

        for method in self.run_methods:
            with self.subTest(method=method):
                case = Using_Default_Loop_TestCase()
                result = getattr(case, method)()

                # assert that the original loop is reset after the test
                self.assertIs(default_loop, asyncio.get_event_loop())

                if result:
                    self.assertTrue(result.wasSuccessful())

            self.assertFalse(default_loop.is_closed())

    def test_forbid_get_event_loop(self):
        default_loop = self.create_default_loop()

        # Changed in python 3.6: get_event_loop() returns the running loop in
        # a callback or a coroutine, so forbid_get_event_loop should only be
        # forbidden where the loop is not running.
        @asynctest.lenient
        class Forbid_get_event_loop_TestCase(asynctest.TestCase):
            forbid_get_event_loop = True

            def runTest(self):
                with self.assertRaises(AssertionError):
                    asyncio.get_event_loop()

        for method in self.run_methods:
            with self.subTest(method=method):
                case = Forbid_get_event_loop_TestCase()
                result = getattr(case, method)()

                # assert that the original loop is reset after the test
                self.assertIs(default_loop, asyncio.get_event_loop())

                if result:
                    self.assertTrue(result.wasSuccessful())

    def test_coroutinefunction_executed(self):
        class CoroutineFunctionTest(asynctest.TestCase):
            ran = False

            @asyncio.coroutine
            def runTest(self):
                self.ran = True
                yield from []

        cases = [CoroutineFunctionTest()]
        if _using_await:
            cases.append(_using_await.CoroutineFunctionTest())

        for method in self.run_methods:
            with self.subTest(method=method):
                for case in cases:
                    with self.subTest(case=case):
                        getattr(case, method)()
                        self.assertTrue(case.ran)

    def test_coroutine_returned_executed(self):
        class CoroutineTest(asynctest.TestCase):
            ran = False

            def runTest(self):
                return asyncio.coroutine(lambda: setattr(self, 'ran', True))()

        cases = [CoroutineTest()]
        if _using_await:
            cases.append(_using_await.CoroutineTest())

        for method in self.run_methods:
            with self.subTest(method=method):
                for case in cases:
                    with self.subTest(case=case):
                        getattr(case, method)()
                        self.assertTrue(case.ran)

    def test_fails_when_future_has_scheduled_calls(self):
        class CruftyTest(asynctest.TestCase):
            @asynctest.fail_on(active_handles=True, unused_loop=False)
            def runTest(instance):
                instance.loop.call_later(5, lambda: None)

        with self.subTest(method='debug'):
            with self.assertRaisesRegex(AssertionError, 'unfinished work'):
                CruftyTest().debug()
        with self.subTest(method='run'):
            result = CruftyTest().run()
            self.assertEqual(1, len(result.failures))

    def test_setup_teardown_may_be_coroutines(self):
        @asynctest.fail_on(unused_loop=False)
        class WithSetupFunction(Test.FooTestCase):
            ran = False

            def setUp(self):
                WithSetupFunction.ran = True

        @asynctest.fail_on(unused_loop=False)
        class WithSetupCoroutine(Test.FooTestCase):
            ran = False

            @asyncio.coroutine
            def setUp(self):
                WithSetupCoroutine.ran = True

        @asynctest.fail_on(unused_loop=False)
        class WithTearDownFunction(Test.FooTestCase):
            ran = False

            def tearDown(self):
                WithTearDownFunction.ran = True

        @asynctest.fail_on(unused_loop=False)
        class WithTearDownCoroutine(Test.FooTestCase):
            ran = False

            def tearDown(self):
                WithTearDownCoroutine.ran = True

        for method in self.run_methods:
            for call_mode in (WithSetupFunction, WithSetupCoroutine,
                              WithTearDownFunction, WithTearDownCoroutine):
                with self.subTest(method=method, case=call_mode.__name__):
                    case = call_mode()
                    call_mode.ran = False

                    getattr(case, method)()
                    self.assertTrue(case.ran)

    def test_cleanup_functions_can_be_coroutines(self):
        cleanup_normal_called = False
        cleanup_normal_called_too_soon = False
        cleanup_coro_called = False

        def cleanup_normal():
            nonlocal cleanup_normal_called
            cleanup_normal_called = True

        @asyncio.coroutine
        def cleanup_coro():
            nonlocal cleanup_coro_called
            cleanup_coro_called = True

        @asynctest.fail_on(unused_loop=False)
        class TestCase(Test.FooTestCase):
            def setUp(self):
                nonlocal cleanup_normal_called, cleanup_normal_called_too_soon
                nonlocal cleanup_coro_called

                cleanup_normal_called = cleanup_coro_called = False

                self.addCleanup(cleanup_normal)
                cleanup_normal_called_too_soon = cleanup_normal_called

                self.addCleanup(cleanup_coro)

        for method in self.run_methods:
            with self.subTest(method=method):
                getattr(TestCase(), method)()
                self.assertTrue(cleanup_normal_called)
                self.assertTrue(cleanup_coro_called)

    def test_loop_uses_TestSelector(self):
        @asynctest.fail_on(unused_loop=False)
        class CheckLoopTest(asynctest.TestCase):
            def runTest(self):
                # TestSelector is used
                self.assertIsInstance(self.loop._selector,
                                      asynctest.selector.TestSelector)

                # And wraps the original selector
                self.assertIsNotNone(self.loop._selector._selector)

        for method in self.run_methods:
            with self.subTest(method=method):
                case = CheckLoopTest()
                outcome = getattr(case, method)()

                if outcome:
                    self.assertTrue(outcome.wasSuccessful())


@unittest.skipIf(sys.platform == "win32", "Tests specific to Unix")
class Test_TestCase_and_ChildWatcher(_TestCase):
    def test_watched_process_is_awaited(self):
        for method in self.run_methods:
            with self.subTest(method=method):
                case = Test.StartWaitProcessTestCase()
                outcome = getattr(case, method)()

                if outcome:
                    self.assertTrue(outcome.wasSuccessful())

    def test_original_watcher_works_outside_loop(self):
        default_loop = self.create_default_loop()

        # check if we can spawn and wait a subprocess before an after a test
        for method in self.run_methods:
            with self.subTest(method=method):
                coro = Test.StartWaitProcessTestCase.start_wait_process(
                    default_loop)
                default_loop.run_until_complete(coro)

                case = Test.StartWaitProcessTestCase()
                outcome = getattr(case, method)()

                if outcome:
                    self.assertTrue(outcome.wasSuccessful())

                coro = Test.StartWaitProcessTestCase.start_wait_process(
                    default_loop)
                default_loop.run_until_complete(coro)


class Test_ClockedTestCase(asynctest.ClockedTestCase):
    took_n_seconds = re.compile('took \d+\.\d{3} seconds')

    @asyncio.coroutine
    def advance(self, seconds):
        try:
            self.loop.set_debug(True)
            with self.assertLogs(level=logging.WARNING) as log:
                yield from self.loop.create_task(super().advance(seconds))

            self.assertTrue(any(filter(self.took_n_seconds.search,
                                       log.output)))
        finally:
            self.loop.set_debug(False)

    @asyncio.coroutine
    def test_advance(self):
        f = asyncio.Future(loop=self.loop)
        g = asyncio.Future(loop=self.loop)
        started_wall_clock = time.monotonic()
        started_loop_clock = self.loop.time()
        self.loop.call_later(1, f.set_result, None)
        self.loop.call_later(2, g.set_result, None)
        self.assertFalse(f.done())
        yield from self.advance(1)
        self.assertTrue(f.done())
        self.assertFalse(g.done())
        yield from self.advance(9)
        self.assertTrue(g)
        finished_wall_clock = time.monotonic()
        finished_loop_clock = self.loop.time()
        self.assertLess(
            finished_wall_clock - started_wall_clock,
            finished_loop_clock - started_loop_clock)

    def test_advance_with_run_until_complete(self):
        f = asyncio.Future(loop=self.loop)
        started_wall_clock = time.monotonic()
        started_loop_clock = self.loop.time()
        self.loop.call_later(1, f.set_result, None)
        self.loop.run_until_complete(self.advance(1))
        self.assertTrue(f.done())
        finished_wall_clock = time.monotonic()
        finished_loop_clock = self.loop.time()
        self.assertLess(
            finished_wall_clock - started_wall_clock,
            finished_loop_clock - started_loop_clock)

    @asyncio.coroutine
    def test_negative_advance(self):
        with self.assertRaisesRegex(ValueError, 'back in time'):
            yield from self.advance(-1)
        self.assertEqual(self.loop.time(), 0)

    @asyncio.coroutine
    def test_callbacks_are_called_on_time(self):
        def record(call_time):
            call_time.append(self.loop.time())

        call_time = []
        self.loop.call_later(0, record, call_time)
        self.loop.call_later(1, record, call_time)
        self.loop.call_later(2, record, call_time)
        self.loop.call_later(5, record, call_time)
        yield from self.advance(3)
        expected = list(range(3))
        self.assertEqual(call_time, expected)
        yield from self.advance(2)
        expected.append(5)
        self.assertEqual(call_time, expected)


class Test_ClockedTestCase_setUp(asynctest.ClockedTestCase):
    def setUp(self):
        pass

    @asyncio.coroutine
    def test_setUp(self):
        yield from self.advance(1)
        self.assertEqual(1, self.loop.time())


class Test_ClockedTestCase_async_setUp(asynctest.ClockedTestCase):
    @asyncio.coroutine
    def setUp(self):
        yield from self.advance(1)

    @asyncio.coroutine
    def test_setUp(self):
        self.assertEqual(1, self.loop.time())


@unittest.mock.patch.dict("asynctest._fail_on.DEFAULTS",
                          values={"foo": False, "bar": True},
                          clear=True)
class Test_fail_on_decorator(unittest.TestCase):
    def checks(self, obj, fatal=True):
        if isinstance(obj, asynctest.TestCase):
            case = obj
        else:
            case = obj.__self__

        try:
            return getattr(obj, asynctest._fail_on._FAIL_ON_ATTR).get_checks(case)
        except AttributeError:
            if fatal:
                self.fail("{!r} not decorated".format(obj))

    def assert_not_decorated(self, obj):
        if self.checks(obj, False) is not None:
            self.fail("{!r} is decorated".format(obj))

    def assert_checks_equal(self, obj, **kwargs):
        self.assertEqual(self.checks(obj), kwargs)

    def assert_checks(self, obj, **kwargs):
        checks = self.checks(obj)
        for check in kwargs:
            if kwargs[check] != checks[check]:
                self.fail("check '{}' {} (expected: {})".format(
                    check,
                    "enabled" if checks[check] else "disabled",
                    "enabled" if kwargs[check] else "disabled")
                )

    def test_check_arguments(self):
        message = "got an unexpected keyword argument 'not_existing'"
        with self.assertRaisesRegex(TypeError, message):
            asynctest.fail_on(foo=True, not_existing=True)

    def test_decorate_method(self):
        class TestCase(asynctest.TestCase):
            @asynctest.fail_on(foo=True)
            def test_foo(self):
                pass

        instance = TestCase()
        self.assert_checks_equal(instance.test_foo, foo=True, bar=True)

    def test_decorate_class(self):
        @asynctest.fail_on(foo=True)
        class TestCase(asynctest.TestCase):
            def test_foo(self):
                pass

            @asynctest.fail_on(bar=False)
            def test_bar(self):
                pass

            @asynctest.fail_on(foo=False, bar=False)
            def test_baz(self):
                pass

        instance = TestCase()
        with self.subTest("class is decorated"):
            self.assert_checks_equal(instance, foo=True, bar=True)

        with self.subTest("method without decoration is not decorated"):
            self.assert_not_decorated(instance.test_foo)

        with self.subTest("method decorated inherits of class params"):
            self.assert_checks_equal(instance.test_bar, foo=True, bar=False)

        with self.subTest("method decorated overrides class params"):
            self.assert_checks_equal(instance.test_baz, foo=False, bar=False)

    def test_decorate_subclass_doesnt_affect_base_class(self):
        class TestCase(asynctest.TestCase):
            pass

        @asynctest.fail_on(foo=True)
        class SubTestCase(TestCase):
            pass

        self.assert_not_decorated(TestCase())
        self.assert_checks(SubTestCase(), foo=True)

    def test_decorate_subclass_inherits_parent_params(self):
        @asynctest.fail_on(foo=True)
        class TestCase(asynctest.TestCase):
            pass

        @asynctest.fail_on(bar=False)
        class SubTestCase(TestCase):
            pass

        @asynctest.fail_on(foo=False, bar=False)
        class OverridingTestCase(TestCase):
            pass

        self.assert_checks_equal(TestCase(), foo=True, bar=True)
        self.assert_checks_equal(SubTestCase(), foo=True, bar=False)
        self.assert_checks_equal(OverridingTestCase(), foo=False, bar=False)

    def test_strict_decorator(self):
        @asynctest.strict
        class TestCase(asynctest.TestCase):
            pass

        self.assert_checks_equal(TestCase(), foo=True, bar=True)

        @asynctest.strict()
        class TestCase(asynctest.TestCase):
            pass

        self.assert_checks_equal(TestCase(), foo=True, bar=True)

    def test_lenient_decorator(self):
        @asynctest.lenient
        class TestCase(asynctest.TestCase):
            pass

        self.assert_checks_equal(TestCase(), foo=False, bar=False)

        @asynctest.lenient()
        class TestCase(asynctest.TestCase):
            pass

        self.assert_checks_equal(TestCase(), foo=False, bar=False)


fail_on_defaults = {"default": True, "optional": False}


@unittest.mock.patch.dict("asynctest._fail_on.DEFAULTS",
                          values=fail_on_defaults, clear=True)
class Test_fail_on(_TestCase):
    def setUp(self):
        self.mocks = {}
        for method in fail_on_defaults:
            mock = self.mocks[method] = unittest.mock.Mock()
            setattr(asynctest._fail_on._fail_on, method, mock)

        self.addCleanup(lambda: [delattr(asynctest._fail_on._fail_on, method)
                                 for method in fail_on_defaults])

    def tearDown(self):
        self.mocks.clear()

    def assert_checked(self, *args):
        for check in args:
            if not self.mocks[check].called:
                self.fail("'{}' unexpectedly unchecked".format(check))

    def assert_not_checked(self, *args):
        for check in args:
            if self.mocks[check].called:
                self.fail("'{}' unexpectedly checked".format(check))

    def test_default_checks(self):
        for method in self.run_methods:
            with self.subTest(method=method):
                case = Test.FooTestCase()
                getattr(case, method)()

                self.assert_checked("default")
                self.assert_not_checked("optional")
                self.mocks["default"].assert_called_with(case)

    def test_checks_on_decorated_class(self):
        @asynctest.fail_on(optional=True)
        class Dummy_TestCase(Test.FooTestCase):
            pass

        for method in self.run_methods:
            with self.subTest(method=method):
                getattr(Dummy_TestCase(), method)()
                self.assert_checked("default", "optional")

    def test_check_after_tearDown(self):
        self.mocks['default'].side_effect = AssertionError

        class Dummy_TestCase(asynctest.TestCase):
            def tearDown(self):
                self.tearDown_called = True

            def runTest(self):
                pass

        with self.subTest(method="debug"):
            case = Dummy_TestCase()
            with self.assertRaises(AssertionError):
                case.debug()

            self.assertTrue(case.tearDown_called)

        case = Dummy_TestCase()
        result = case.run()
        self.assertEqual(1, len(result.failures))
        self.assertTrue(case.tearDown_called)

    def test_before_test_called_before_user_setup(self):
        mock = unittest.mock.Mock()
        setattr(asynctest._fail_on._fail_on, "before_test_default", mock)
        self.addCleanup(delattr, asynctest._fail_on._fail_on,
                        "before_test_default")

        class TestCase(asynctest.TestCase):
            def setUp(self):
                self.assertTrue(mock.called)

            def runTest(self):
                pass

        for method in self.run_methods:
            with self.subTest(method=method):
                case = Test.FooTestCase()
                getattr(case, method)()

                self.assert_checked("default")
                self.assertTrue(mock.called)
                mock.assert_called_with(case)

    def test_non_existing_before_test_wont_fail(self):
        # set something not callable for default, nothing for optional, the
        # test must not fail
        setattr(asynctest._fail_on._fail_on, "before_test_default", None)
        self.addCleanup(delattr, asynctest._fail_on._fail_on,
                        "before_test_default")

        @asynctest.fail_on(default=True, optional=True)
        class TestCase(asynctest.TestCase):
            def runTest(self):
                pass

        for method in self.run_methods:
            with self.subTest(method=method):
                getattr(TestCase(), method)()
                self.assert_checked("default", "optional")

    def test_before_test_called_for_enabled_checks_only(self):
        for method in map(lambda m: "before_test_" + m, fail_on_defaults):
            mock = self.mocks[method] = unittest.mock.Mock()
            setattr(asynctest._fail_on._fail_on, method, mock)

        self.addCleanup(lambda: [delattr(asynctest._fail_on._fail_on,
                                         "before_test_" + method)
                                 for method in fail_on_defaults])

        for method in self.run_methods:
            with self.subTest(method=method):
                getattr(Test.FooTestCase(), method)()
                self.assertTrue(self.mocks["before_test_default"].called)
                self.assertFalse(self.mocks["before_test_optional"].called)


@unittest.mock.patch.dict(
    "asynctest._fail_on.DEFAULTS", clear=True,
    unused_loop=asynctest._fail_on.DEFAULTS['unused_loop'])
class Test_fail_on_unused_loop(_TestCase):
    @asynctest.fail_on(unused_loop=True)
    class WithCheckTestCase(Test.FooTestCase):
        pass

    def test_fails_when_loop_didnt_run(self):
        with self.assertRaisesRegex(AssertionError,
                                    'Loop did not run during the test'):
            self.WithCheckTestCase().debug()

        result = self.WithCheckTestCase().run()
        self.assertEqual(1, len(result.failures))

    def test_fails_when_loop_didnt_run_using_default_loop(self):
        class TestCase(self.WithCheckTestCase):
            use_default_loop = True

        default_loop = self.create_default_loop()

        with self.assertRaisesRegex(AssertionError,
                                    'Loop did not run during the test'):
            TestCase().debug()

        result = TestCase().run()
        self.assertEqual(1, len(result.failures))

        default_loop.run_until_complete(asyncio.sleep(0, loop=default_loop))

        with self.assertRaisesRegex(AssertionError,
                                    'Loop did not run during the test'):
            TestCase().debug()

        default_loop.run_until_complete(asyncio.sleep(0, loop=default_loop))

        result = TestCase().run()
        self.assertEqual(1, len(result.failures))

    def test_passes_when_ignore_loop_or_loop_run(self):
        @asynctest.fail_on(unused_loop=False)
        class IgnoreLoopClassTest(Test.FooTestCase):
            pass

        @asynctest.fail_on(unused_loop=True)
        class WithCoroutineTest(asynctest.TestCase):
            @asyncio.coroutine
            def runTest(self):
                yield from []

        @asynctest.fail_on(unused_loop=True)
        class WithFunctionCallingLoopTest(asynctest.TestCase):
            def runTest(self):
                fut = asyncio.Future()
                self.loop.call_soon(fut.set_result, None)
                self.loop.run_until_complete(fut)

        for test in (IgnoreLoopClassTest, WithCoroutineTest,
                     WithFunctionCallingLoopTest):
            with self.subTest(test=test):
                test().debug()

                result = test().run()
                self.assertEqual(0, len(result.failures))

    def test_fails_when_loop_ran_only_during_setup(self):
        for test_use_default_loop in (False, True):
            with self.subTest(use_default_loop=test_use_default_loop):
                if test_use_default_loop:
                    self.create_default_loop()

                class TestCase(self.WithCheckTestCase):
                    use_default_loop = test_use_default_loop

                    def setUp(self):
                        self.loop.run_until_complete(
                            asyncio.sleep(0, loop=self.loop))

                with self.assertRaisesRegex(
                        AssertionError, 'Loop did not run during the test'):
                    TestCase().debug()

                result = TestCase().run()
                self.assertEqual(1, len(result.failures))

    def test_fails_when_loop_ran_only_during_cleanup(self):
        for test_use_default_loop in (False, True):
            with self.subTest(use_default_loop=test_use_default_loop):
                if test_use_default_loop:
                    self.create_default_loop()

                class TestCase(self.WithCheckTestCase):
                    use_default_loop = test_use_default_loop

                    def setUp(self):
                        self.addCleanup(asyncio.coroutine(lambda: None))

                with self.assertRaisesRegex(
                        AssertionError, 'Loop did not run during the test'):
                    TestCase().debug()

                result = TestCase().run()
                self.assertEqual(1, len(result.failures))


class Test_assertAsyncRaises(asynctest.TestCase):
    class CustomException(Exception):
        def __str__(self):
            return "CustomException Message"

    @property
    def raising(self):
        awaitable = asyncio.Future()
        awaitable.set_exception(self.CustomException())
        return awaitable

    @property
    def finishing(self):
        awaitable = asyncio.Future()
        awaitable.set_result(None)
        return awaitable

    @asyncio.coroutine
    def test_assertAsyncRaises(self):
        with self.subTest("exception raised"):
            yield from self.assertAsyncRaises(self.CustomException,
                                              self.raising)

        with self.subTest("no exception raised"):
            with self.assertRaises(AssertionError):
                yield from self.assertAsyncRaises(self.CustomException,
                                                  self.finishing)

    @asyncio.coroutine
    def test_assertAsyncRaisesRegex(self):
        regex = "CustomException"
        with self.subTest("exception raised"):
            yield from self.assertAsyncRaisesRegex(self.CustomException,
                                                   regex, self.raising)

        with self.subTest("no exception raised"):
            with self.assertRaises(AssertionError):
                yield from self.assertAsyncRaisesRegex(self.CustomException,
                                                       regex, self.finishing)


class Test_assertAsyncWarns(asynctest.TestCase):
    class CustomWarning(Warning):
        pass

    @classmethod
    @asyncio.coroutine
    def warns(cls):
        import warnings
        warnings.warn("asynctest warning message", cls.CustomWarning)

    @staticmethod
    def doesnt_warn():
        awaitable = asyncio.Future()
        awaitable.set_result(None)
        return awaitable

    @asyncio.coroutine
    def test_assertAsyncWarns(self):
        with self.subTest("warning triggered"):
            yield from self.assertAsyncWarns(self.CustomWarning, self.warns())

        with self.subTest("warning not triggered"):
            with self.assertRaises(AssertionError):
                yield from self.assertAsyncWarns(self.CustomWarning,
                                                 self.doesnt_warn())

    @asyncio.coroutine
    def test_assertAsyncWarnsRegex(self):
        regex = "asynctest warning"
        with self.subTest("warning triggered"):
            yield from self.assertAsyncWarnsRegex(self.CustomWarning, regex,
                                                  self.warns())

        with self.subTest("warning not triggered"):
            with self.assertRaises(AssertionError):
                yield from self.assertAsyncWarnsRegex(
                    self.CustomWarning, regex, self.doesnt_warn())


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

import asyncio
import functools
import inspect
import unittest
import sys

import asynctest

from .utils import run_coroutine

if sys.version_info >= (3, 5):
    from . import test_mock_await as _using_await
else:
    _using_await = None


class Test:
    @asyncio.coroutine
    def a_coroutine(self):
        pass

    def a_function(self):
        pass

    def is_patched(self):
        return False

    def second_is_patched(self):
        return False

    a_dict = {'is_patched': False, 'second_is_patched': False}
    a_second_dict = {'is_patched': False}

    @asyncio.coroutine
    def a_coroutine_with_args(self, arg, arg2):
        return None


class ProbeException(Exception):
    pass


if _using_await:
    Test = _using_await.patch_Test_Class(Test)


patch_is_patched = functools.partial(asynctest.mock.patch,
                                     'test.test_mock.Test.is_patched',
                                     new=lambda self: True)

patch_second_is_patched = functools.partial(
    asynctest.mock.patch, 'test.test_mock.Test.second_is_patched',
    new=lambda self: True)

patch_dict_is_patched = functools.partial(
    asynctest.mock.patch.dict, 'test.test_mock.Test.a_dict',
    values={"is_patched": True})

patch_dict_second_is_patched = functools.partial(
    asynctest.mock.patch.dict, 'test.test_mock.Test.a_dict',
    values={"second_is_patched": True})

patch_dict_second_dict_is_patched = functools.partial(
    asynctest.mock.patch.dict, 'test.test_mock.Test.a_second_dict',
    values={"is_patched": True})


def inject_class(obj):
    # Decorate _Test_* mixin classes so we can retrieve the mock class to test
    # with the last argument of the test function ("klass").
    if isinstance(obj, type):
        for attr_name in dir(obj):
            attr = getattr(obj, attr_name)
            if callable(attr) and attr_name.startswith('test_'):
                setattr(obj, attr_name, inject_class(attr))

        return obj
    else:
        @functools.wraps(obj)
        def wrapper(self):
            return obj(self, getattr(asynctest, self.class_to_test))

        return wrapper


@inject_class
class _Test_iscoroutinefunction:
    # Ensure that an instance of this mock type is seen as a coroutine function
    def test_asyncio_iscoroutinefunction(self, klass):
        with self.subTest(is_coroutine=False):
            mock = klass(is_coroutine=False)
            self.assertFalse(asyncio.iscoroutinefunction(mock))

        with self.subTest(is_coroutine=False):
            mock = klass(is_coroutine=True)
            self.assertTrue(asyncio.iscoroutinefunction(mock))


@inject_class
class _Test_is_coroutine_property:
    # Ensure an instance offers an is_coroutine property
    def test_is_coroutine_property(self, klass):
        mock = klass()
        self.assertFalse(mock.is_coroutine)

        mock.is_coroutine = True
        self.assertTrue(mock.is_coroutine)

        mock = klass(is_coroutine=True)
        self.assertTrue(mock.is_coroutine)


@inject_class
class _Test_subclass:
    # Ensure that the tested class is also a subclass of its counterpart in
    # the standard module unittest.mock
    def test_subclass(self, klass):
        unittest_klass = getattr(unittest.mock, self.class_to_test)

        self.assertTrue(issubclass(klass, unittest_klass))
        self.assertTrue(isinstance(klass(), unittest_klass))


@inject_class
class _Test_called_coroutine:
    # Ensure that an object mocking as a coroutine works
    def test_returns_coroutine(self, klass):
        mock = klass()

        coro = mock()
        # Suppress debug warning about non-running coroutine
        if isinstance(coro, asyncio.coroutines.CoroWrapper):
            coro.gen = None

        self.assertTrue(asyncio.iscoroutine(coro))

    def test_returns_coroutine_from_return_value(self, klass):
        mock = klass()
        mock.return_value = 'ProbeValue'

        self.assertEqual('ProbeValue', mock.return_value)
        self.assertEqual(mock.return_value, run_coroutine(mock()))

    def test_returns_coroutine_with_return_value_being_a_coroutine(self, klass):
        mock = klass()
        coroutine = asyncio.coroutine(lambda: 'ProbeValue')
        mock.return_value = coroutine()

        self.assertEqual('ProbeValue', run_coroutine(mock()))

    def test_returns_coroutine_from_side_effect(self, klass):
        mock = klass()
        mock.side_effect = lambda: 'ProbeValue'

        self.assertEqual('ProbeValue', run_coroutine(mock()))

    def test_returns_coroutine_from_side_effect_being_a_coroutine(self, klass):
        mock = klass()
        mock.side_effect = asyncio.coroutine(lambda: 'ProbeValue')

        self.assertEqual('ProbeValue', run_coroutine(mock()))

    def test_exception_side_effect_raises_in_coroutine(self, klass):
        mock = klass()
        mock.side_effect = ProbeException

        coroutine = mock()
        with self.assertRaises(ProbeException):
            run_coroutine(coroutine)

    def test_returns_coroutine_from_side_effect_being_an_iterable(self, klass):
        mock = klass()
        side_effect = ['Probe1', 'Probe2', 'Probe3']
        mock.side_effect = side_effect

        for expected in side_effect:
            self.assertEqual(expected, run_coroutine(mock()))

        with self.assertRaises(StopIteration):
            mock()


@inject_class
class _Test_Spec_Spec_Set_Returns_Coroutine_Mock:
    # Ensure that when a mock is configured with spec or spec_set, coroutines
    # are detected and mocked correctly
    def test_mock_returns_coroutine_according_to_spec(self, klass):
        spec = Test()

        for attr in ('spec', 'spec_set', ):
            with self.subTest(spec_type=attr):
                mock = klass(**{attr: spec})

                self.assertIsInstance(mock.a_function, (asynctest.Mock, asynctest.MagicMock))
                self.assertNotIsInstance(mock.a_function, asynctest.CoroutineMock)
                self.assertIsInstance(mock.a_coroutine, asynctest.CoroutineMock)
                mock.a_coroutine.return_value = "PROBE"
                self.assertEqual("PROBE", run_coroutine(mock.a_coroutine()))

                if _using_await:
                    self.assertIsInstance(mock.an_async_coroutine, asynctest.CoroutineMock)
                    mock.an_async_coroutine.return_value = "PROBE"
                    self.assertEqual("PROBE", run_coroutine(mock.an_async_coroutine()))

    # Ensure the name of the mock is correctly set, tests bug #49.
    def test_mock_has_correct_name(self, klass):
        spec = Test()

        for attr in ('spec', 'spec_set', ):
            with self.subTest(spec_type=attr):
                mock = klass(**{attr: spec})

                self.assertIn("{}='{}".format(attr, "Test"), repr(mock))
                self.assertIn("name='mock.a_coroutine'", repr(mock.a_coroutine))
                mock.a_coroutine()  # is a generator, not a Mock with a repr
                self.assertIn("name='mock.a_function()'", repr(mock.a_function()))
                self.assertEqual("call.a_coroutine()", repr(mock.mock_calls[0]))
                self.assertEqual("call.a_function()", repr(mock.mock_calls[1]))


@inject_class
class _Test_Spec_Spec_Set_Is_Function:
    def test_mock_is_not_coroutine_when_spec_is_function(self, klass):
        spec = Test.a_function

        for attr in ('spec', 'spec_set', ):
            with self.subTest(spec_type=attr):
                mock = klass(**{attr: spec})
                self.assertFalse(asyncio.iscoroutinefunction(mock))
                if hasattr(inspect, "iscoroutinefunction"):
                    self.assertFalse(inspect.iscoroutinefunction(mock))


@inject_class
class _Test_Future:
    # Ensure that a mocked Future is detected as a future
    def test_mock_a_future_is_a_future(self, klass):
        mock = klass(asyncio.Future())
        self.assertIsInstance(mock, asyncio.Future)

    def test_mock_from_create_future(self, klass):
        loop = asyncio.new_event_loop()

        try:
            if not (hasattr(loop, "create_future") and
                    hasattr(asyncio, "isfuture")):
                return

            mock = klass(loop.create_future())
            self.assertTrue(asyncio.isfuture(mock))
        finally:
            loop.close()


# Import mixins based on the support of async/await keywords
if _using_await:
    _Test_Mock_Of_Async_Magic_Methods = inject_class(
        _using_await._Test_Mock_Of_Async_Magic_Methods)
else:
    class _Test_Mock_Of_Async_Magic_Methods:
        pass


class Test_NonCallabableMock(unittest.TestCase, _Test_subclass,
                             _Test_iscoroutinefunction,
                             _Test_is_coroutine_property,
                             _Test_Spec_Spec_Set_Returns_Coroutine_Mock,
                             _Test_Spec_Spec_Set_Is_Function,
                             _Test_Future):
    class_to_test = 'NonCallableMock'


class Test_NonCallableMagicMock(unittest.TestCase, _Test_subclass,
                                _Test_iscoroutinefunction,
                                _Test_is_coroutine_property,
                                _Test_Spec_Spec_Set_Returns_Coroutine_Mock,
                                _Test_Spec_Spec_Set_Is_Function,
                                _Test_Future,
                                _Test_Mock_Of_Async_Magic_Methods):
    class_to_test = 'NonCallableMagicMock'


class Test_Mock(unittest.TestCase, _Test_subclass,
                _Test_Spec_Spec_Set_Returns_Coroutine_Mock,
                _Test_Spec_Spec_Set_Is_Function,
                _Test_Future):
    class_to_test = 'Mock'


class Test_MagicMock(unittest.TestCase, _Test_subclass,
                     _Test_Spec_Spec_Set_Returns_Coroutine_Mock,
                     _Test_Spec_Spec_Set_Is_Function,
                     _Test_Future, _Test_Mock_Of_Async_Magic_Methods):
    class_to_test = 'MagicMock'


class Test_CoroutineMock(unittest.TestCase, _Test_called_coroutine,
                         _Test_Spec_Spec_Set_Returns_Coroutine_Mock):
    class_to_test = 'CoroutineMock'

    def test_asyncio_iscoroutinefunction(self):
        mock = asynctest.mock.CoroutineMock()
        self.assertTrue(asyncio.iscoroutinefunction(mock))

    def test_called_CoroutineMock_returns_MagicMock(self):
        mock = asynctest.mock.CoroutineMock()
        self.assertIsInstance(run_coroutine(mock()), asynctest.mock.MagicMock)


class Test_CoroutineMock_awaited(asynctest.TestCase):
    @asynctest.fail_on(unused_loop=False)
    def test_awaited_delays_creation_of_condition(self):
        mock = asynctest.mock.CoroutineMock()
        self.assertIsNone(mock.awaited._condition)
        mock()
        self.assertIsNone(mock.awaited._condition)
        run_coroutine(mock())
        self.assertIsNotNone(mock.awaited._condition)

        @asyncio.coroutine
        def f():
            yield from mock.awaited._condition.acquire()
            yield from mock.awaited._condition.acquire()

        # Condition belongs to a loop created by run_coroutine, not current loop.
        with self.assertRaises(RuntimeError):
            run_coroutine(f())

    @asyncio.coroutine
    def test_awaited_CoroutineMock_sets_awaited(self):
        mock = asynctest.mock.CoroutineMock()
        yield from mock()
        self.assertTrue(mock.awaited)

        mock.reset_mock()
        self.assertFalse(mock.awaited)

        @asyncio.coroutine
        def side_effect():
            raise RuntimeError()

        mock = asynctest.mock.CoroutineMock(side_effect=side_effect)

        with self.assertRaises(RuntimeError):
            yield from mock()

    @asyncio.coroutine
    def test_awaited_CoroutineMock_counts(self):
        mock = asynctest.mock.CoroutineMock()
        yield from mock()
        yield from mock()
        self.assertEqual(mock.await_count, 2)

        mock.reset_mock()
        self.assertEqual(mock.await_count, 0)

        @asyncio.coroutine
        def side_effect():
            raise RuntimeError()

        mock = asynctest.mock.CoroutineMock(side_effect=side_effect)

        with self.assertRaises(RuntimeError):
            yield from mock()

        self.assertEqual(mock.await_count, 1)

    @asyncio.coroutine
    def test_awaited_from_autospec_mock(self):
        mock = asynctest.mock.create_autospec(Test)
        self.assertFalse(mock.a_coroutine.awaited)
        self.assertEqual(0, mock.a_coroutine.await_count)

        yield from mock.a_coroutine()

        self.assertTrue(mock.a_coroutine.awaited)
        self.assertEqual(1, mock.a_coroutine.await_count)

    @asyncio.coroutine
    def test_awaited_wait(self):
        mock = asynctest.mock.CoroutineMock()
        t = asyncio.ensure_future(mock.awaited.wait())
        yield from mock()
        yield from mock()
        yield from t

        mock.reset_mock()
        t = asyncio.ensure_future(mock.awaited.wait(skip=1))
        yield from mock()
        self.assertFalse(t.done())
        yield from mock()
        yield from t

    @asyncio.coroutine
    def test_awaited_wait_next(self):
        mock = asynctest.mock.CoroutineMock()
        yield from mock()
        t = asyncio.ensure_future(mock.awaited.wait_next())
        yield from asyncio.sleep(0.01)
        self.assertFalse(t.done())
        yield from mock()
        yield from t

        mock.reset_mock()
        yield from mock()
        t = asyncio.ensure_future(mock.awaited.wait_next(skip=1))
        yield from asyncio.sleep(0.01)
        yield from mock()
        self.assertFalse(t.done())
        yield from mock()
        yield from t

    @asyncio.coroutine
    def test_await_args(self):
        with self.subTest('in order'):
            mock = asynctest.mock.CoroutineMock()
            t1 = mock('foo')
            t2 = mock('bar')
            yield from t1
            yield from t2
            self.assertEqual(mock.await_args, asynctest.call('bar'))

        with self.subTest('out of order'):
            mock = asynctest.mock.CoroutineMock()
            t1 = mock('foo')
            t2 = mock('bar')
            yield from t2
            yield from t1
            self.assertEqual(mock.await_args, asynctest.call('foo'))

    @asyncio.coroutine
    def test_await_args_list(self):
        with self.subTest('in order'):
            mock = asynctest.mock.CoroutineMock()
            t1 = mock('foo')
            t2 = mock('bar')
            yield from t1
            yield from t2
            self.assertEqual(mock.await_args_list, [asynctest.call('foo'), asynctest.call('bar')])

        with self.subTest('out of order'):
            mock = asynctest.mock.CoroutineMock()
            t1 = mock('foo')
            t2 = mock('bar')
            yield from t2
            yield from t1
            self.assertEqual(mock.await_args_list, [asynctest.call('bar'), asynctest.call('foo')])

    @asyncio.coroutine
    def test_assert_awaited(self):
        mock = asynctest.mock.CoroutineMock()

        with self.assertRaises(AssertionError):
            mock.assert_awaited()

        yield from mock()
        mock.assert_awaited()

    @asyncio.coroutine
    def test_assert_awaited_once(self):
        mock = asynctest.mock.CoroutineMock()

        with self.assertRaises(AssertionError):
            mock.assert_awaited_once()

        yield from mock()
        mock.assert_awaited_once()

        yield from mock()
        with self.assertRaises(AssertionError):
            mock.assert_awaited_once()

    @asyncio.coroutine
    def test_assert_awaited_with(self):
        mock = asynctest.mock.CoroutineMock()

        with self.assertRaises(AssertionError):
            mock.assert_awaited_with('foo')

        yield from mock('foo')
        mock.assert_awaited_with('foo')

        yield from mock('bar')
        with self.assertRaises(AssertionError):
            mock.assert_awaited_with('foo')

    @asyncio.coroutine
    def test_assert_awaited_once_with(self):
        mock = asynctest.mock.CoroutineMock()

        with self.assertRaises(AssertionError):
            mock.assert_awaited_once_with('foo')

        yield from mock('foo')
        mock.assert_awaited_once_with('foo')

        yield from mock('foo')
        with self.assertRaises(AssertionError):
            mock.assert_awaited_once_with('foo')

    @asyncio.coroutine
    def test_assert_any_wait(self):
        mock = asynctest.mock.CoroutineMock()

        with self.assertRaises(AssertionError):
            mock.assert_any_await('bar')

        yield from mock('foo')
        with self.assertRaises(AssertionError):
            mock.assert_any_await('bar')

        yield from mock('bar')
        mock.assert_any_await('bar')

        yield from mock('baz')
        mock.assert_any_await('bar')

    @asyncio.coroutine
    def test_assert_has_awaits(self):
        calls = [asynctest.call('bar'), asynctest.call('baz')]

        with self.subTest('any_order=False'):
            mock = asynctest.mock.CoroutineMock()

            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls)

            yield from mock('foo')
            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls)

            yield from mock('bar')
            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls)

            yield from mock('baz')
            mock.assert_has_awaits(calls)

            yield from mock('qux')
            mock.assert_has_awaits(calls)

        with self.subTest('any_order=True'):
            mock = asynctest.mock.CoroutineMock()

            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls, any_order=True)

            yield from mock('baz')
            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls, any_order=True)

            yield from mock('foo')
            with self.assertRaises(AssertionError):
                mock.assert_has_awaits(calls, any_order=True)

            yield from mock('bar')
            mock.assert_has_awaits(calls, any_order=True)

            yield from mock('qux')
            mock.assert_has_awaits(calls, any_order=True)

    @asyncio.coroutine
    def test_assert_not_awaited(self):
        mock = asynctest.mock.CoroutineMock()

        mock.assert_not_awaited()

        yield from mock()
        with self.assertRaises(AssertionError):
            mock.assert_not_awaited()


class TestMockInheritanceModel(unittest.TestCase):
    to_test = {
        'NonCallableMagicMock': 'NonCallableMock',
        'Mock': 'NonCallableMock',
        'MagicMock': 'Mock',
        'CoroutineMock': 'Mock',
    }

    def test_Mock_is_not_CoroutineMock(self):
        self.assertNotIsInstance(asynctest.mock.Mock(), asynctest.mock.CoroutineMock)

    def test_MagicMock_is_not_CoroutineMock(self):
        self.assertNotIsInstance(asynctest.mock.MagicMock(), asynctest.mock.CoroutineMock)

    @staticmethod
    def make_inheritance_test(child, parent):
        def test(self):
            # Works in the common case
            self.assertIsInstance(getattr(asynctest.mock, child)(),
                                  getattr(asynctest.mock, parent))

            # Works with a custom spec
            self.assertIsInstance(getattr(asynctest.mock, child)(Test()),
                                  getattr(asynctest.mock, parent))

        return test


for child, parent in TestMockInheritanceModel.to_test.items():
    setattr(TestMockInheritanceModel,
            'test_{}_inherits_from_{}'.format(child, parent),
            TestMockInheritanceModel.make_inheritance_test(child, parent))

#
# mock_open()
#


class Test_mock_open(unittest.TestCase):
    def test_MagicMock_returned_by_default(self):
        self.assertIsInstance(asynctest.mock_open(), asynctest.MagicMock)

#
# Test patches
#


class Test_patch(unittest.TestCase):
    def test_patch_as_context_manager_uses_MagicMock(self):
        with asynctest.mock.patch('test.test_mock.Test') as mock:
            self.assertIsInstance(mock, asynctest.mock.MagicMock)

        with asynctest.mock.patch('test.test_mock.Test.a_function') as mock:
            self.assertIsInstance(mock, asynctest.mock.MagicMock)

    def test_patch_as_decorator_uses_MagicMock(self):
        called = []

        @asynctest.mock.patch('test.test_mock.Test')
        def test_mock_class(mock):
            self.assertIsInstance(mock, asynctest.mock.MagicMock)
            called.append("test_mock_class")

        @asynctest.mock.patch('test.test_mock.Test.a_function')
        def test_mock_function(mock):
            self.assertIsInstance(mock, asynctest.mock.MagicMock)
            called.append("test_mock_function")

        test_mock_class()
        test_mock_function()

        self.assertIn("test_mock_class", called)
        self.assertIn("test_mock_function", called)

    def test_patch_as_decorator_uses_CoroutineMock_on_coroutine_function(self):
        called = False

        @asynctest.mock.patch('test.test_mock.Test.a_coroutine')
        def test_mock_coroutine(mock):
            nonlocal called
            self.assertIsInstance(mock, asynctest.mock.CoroutineMock)
            called = True

        test_mock_coroutine()
        self.assertTrue(called)

    def test_patch_as_context_manager_uses_CoroutineMock_on_coroutine_function(self):
        with asynctest.mock.patch('test.test_mock.Test.a_coroutine'):
            import test.test_mock
            self.assertIsInstance(test.test_mock.Test.a_coroutine,
                                  asynctest.mock.CoroutineMock)

    if _using_await:
        def test_patch_as_context_manager_uses_CoroutineMock_on_async_coroutine_function(self):
            with asynctest.mock.patch('test.test_mock.Test.an_async_coroutine') as mock:
                self.assertIsInstance(mock, asynctest.mock.CoroutineMock)

        def test_patch_as_decorator_uses_CoroutineMock_on__async_coroutine_function(self):
            @asynctest.mock.patch('test.test_mock.Test.an_async_coroutine')
            def test_mock_coroutine(mock):
                self.assertIsInstance(mock, asynctest.mock.CoroutineMock)

            test_mock_coroutine()

    def test_patch_is_enabled_when_running_decorated_coroutine(self):
        @asyncio.coroutine
        def a_coroutine():
            import test.test_mock
            return test.test_mock.Test().is_patched()

        coroutines = [a_coroutine]
        if _using_await:
            coroutines.append(_using_await.transform(a_coroutine))

        for coroutine in coroutines:
            with self.subTest(coroutine=coroutine):
                self.assertTrue(run_coroutine(patch_is_patched()(coroutine)()))

    def test_patch_is_enabled_when_running_decorated_function(self):
        @patch_is_patched()
        def a_function():
            import test.test_mock
            return test.test_mock.Test().is_patched()

        self.assertTrue(a_function())


class Test_patch_decorator_coroutine_or_generator(unittest.TestCase):
    def test_coroutine_type_when_patched(self):
        @asyncio.coroutine
        def a_coroutine():
            pass

        a_patched_coroutine = patch_is_patched()(a_coroutine)

        self.assertEqual(asyncio.iscoroutinefunction(a_patched_coroutine),
                         asyncio.iscoroutinefunction(a_coroutine))
        self.assertEqual(inspect.isgeneratorfunction(a_patched_coroutine),
                         inspect.isgeneratorfunction(a_coroutine))
        coro = a_coroutine()
        patched_coro = a_patched_coroutine()
        try:
            self.assertEqual(asyncio.iscoroutine(patched_coro),
                             asyncio.iscoroutine(coro))
        finally:
            run_coroutine(coro)
            run_coroutine(patched_coro)

        if not _using_await:
            return

        a_coroutine = _using_await.transform(a_coroutine)
        a_patched_coroutine = patch_is_patched()(a_coroutine)
        self.assertEqual(asyncio.iscoroutinefunction(a_patched_coroutine),
                         asyncio.iscoroutinefunction(a_coroutine))
        coro = a_coroutine()
        patched_coro = a_patched_coroutine()
        try:
            self.assertEqual(asyncio.iscoroutine(patched_coro),
                             asyncio.iscoroutine(coro))
        finally:
            run_coroutine(coro)
            run_coroutine(patched_coro)

    def test_generator_arg_is_default_mock(self):
        @asynctest.mock.patch('test.test_mock.Test')
        def a_generator(mock):
            self.assertIsInstance(mock, asynctest.mock.Mock)
            yield
            import test.test_mock
            self.assertIs(mock, test.test_mock.Test)

        for _ in a_generator():
            pass

    def test_coroutine_arg_is_default_mock(self):
        @asyncio.coroutine
        def tester(coroutine_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(fut.set_result, None)
            before, after = yield from coroutine_function(fut)
            self.assertTrue(before)
            self.assertTrue(after)

        def is_instance_of_mock(obj):
            return isinstance(obj, asynctest.mock.Mock)

        def is_same_mock(obj):
            import test.test_mock
            return obj is test.test_mock.Test

        with self.subTest("old style coroutine"):
            @asynctest.mock.patch('test.test_mock.Test')
            def a_coroutine(fut, mock):
                before = is_instance_of_mock(mock)
                yield from fut
                after = is_same_mock(mock)
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                is_instance_of_mock, is_same_mock)
            a_new_style_coroutine = asynctest.mock.patch(
                'test.test_mock.Test')(a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))


class Test_patch_object(unittest.TestCase):
    def test_patch_with_MagicMock(self):
        with asynctest.mock.patch.object(Test(), 'a_function') as mock:
            self.assertIsInstance(mock, asynctest.mock.MagicMock)

        obj = Test()
        obj.test = Test()
        with asynctest.mock.patch.object(obj, 'test') as mock:
            self.assertIsInstance(mock, asynctest.mock.MagicMock)

    def test_patch_coroutine_function_with_CoroutineMock(self):
        with asynctest.mock.patch.object(Test(), 'a_coroutine') as mock:
            self.assertIsInstance(mock, asynctest.mock.CoroutineMock)

        if _using_await:
            with asynctest.mock.patch.object(Test(), 'an_async_coroutine') as mock:
                self.assertIsInstance(mock, asynctest.mock.CoroutineMock)

    def test_patch_decorates_coroutine(self):
        obj = Test()

        patch = functools.partial(asynctest.mock.patch.object,
                                  obj, 'is_patched', new=lambda: True)

        @asyncio.coroutine
        def a_coroutine():
            return obj.is_patched()

        coroutines = [a_coroutine]
        if _using_await:
            coroutines.append(_using_await.transform(a_coroutine))

        for coroutine in coroutines:
            with self.subTest(coroutine=coroutine):
                self.assertTrue(run_coroutine(patch()(coroutine)()))


class Test_patch_multiple(unittest.TestCase):
    def test_patch_with_MagicMock(self):
        default = asynctest.mock.DEFAULT
        with asynctest.mock.patch.multiple('test.test_mock', Test=default):
            import test.test_mock
            self.assertIsInstance(test.test_mock.Test, asynctest.mock.MagicMock)

    def test_patch_coroutine_function_with_CoroutineMock(self):
        default = asynctest.mock.DEFAULT

        also_patch = {}
        if _using_await:
            also_patch['an_async_coroutine'] = default

        with asynctest.mock.patch.multiple('test.test_mock.Test',
                                           a_function=default,
                                           a_coroutine=default,
                                           **also_patch):
            import test.test_mock
            obj = test.test_mock.Test()
            self.assertIsInstance(obj.a_function, asynctest.mock.MagicMock)
            self.assertIsInstance(obj.a_coroutine, asynctest.mock.CoroutineMock)

            if _using_await:
                self.assertIsInstance(obj.an_async_coroutine,
                                      asynctest.mock.CoroutineMock)

    def test_patch_decorates_coroutine(self):
        patch = functools.partial(asynctest.mock.patch.multiple,
                                  'test.test_mock.Test',
                                  is_patched=lambda self: True)

        @asyncio.coroutine
        def a_coroutine():
            import test.test_mock
            return test.test_mock.Test().is_patched()

        coroutines = [a_coroutine]
        if _using_await:
            coroutines.append(_using_await.transform(a_coroutine))

        for coroutine in coroutines:
            with self.subTest(coroutine=coroutine):
                self.assertTrue(run_coroutine(patch()(coroutine)()))


class Test_patch_dict(unittest.TestCase):
    def test_patch_decorates_coroutine(self):
        @asyncio.coroutine
        def a_coroutine():
            import test.test_mock
            return test.test_mock.Test().a_dict['is_patched']

        coroutines = [a_coroutine]

        if _using_await:
            coroutines.append(_using_await.transform(a_coroutine))

        for coroutine in coroutines:
            with self.subTest(coroutine=coroutine):
                self.assertTrue(run_coroutine(patch_dict_is_patched()(coroutine)()))

    def test_patch_decorates_function(self):
        @patch_dict_is_patched()
        def a_function():
            import test.test_mock
            return test.test_mock.Test().a_dict['is_patched']

        self.assertTrue(a_function())

    def test_patch_decorates_class(self):
        import test.test_mock

        @patch_dict_is_patched()
        class Patched:
            @asyncio.coroutine
            def test_a_coroutine(self):
                return test.test_mock.Test().a_dict['is_patched']

            def test_a_function(self):
                return test.test_mock.Test().a_dict['is_patched']

        instance = Patched()
        self.assertFalse(test.test_mock.Test().a_dict['is_patched'])
        self.assertTrue(instance.test_a_function())
        self.assertFalse(test.test_mock.Test().a_dict['is_patched'])
        self.assertTrue(run_coroutine(instance.test_a_coroutine()))
        self.assertFalse(test.test_mock.Test().a_dict['is_patched'])


class Test_patch_autospec(unittest.TestCase):
    test_class_path = "{}.Test".format(__name__)

    def test_autospec_coroutine(self):
        called = False

        @asynctest.mock.patch(self.test_class_path, autospec=True)
        def patched(mock):
            nonlocal called
            called = True
            self.assertIsInstance(mock.a_coroutine,
                                  asynctest.mock.CoroutineMock)

            self.assertIsInstance(mock().a_coroutine,
                                  asynctest.mock.CoroutineMock)

            self.assertIsInstance(mock.a_function, asynctest.mock.Mock)
            self.assertIsInstance(mock().a_function, asynctest.mock.Mock)

            if _using_await:
                self.assertIsInstance(mock.an_async_coroutine,
                                      asynctest.mock.CoroutineMock)
                self.assertIsInstance(mock().an_async_coroutine,
                                      asynctest.mock.CoroutineMock)

        patched()
        self.assertTrue(called)

    def test_patch_autospec_with_patches_on_top(self):
        called = False

        @asynctest.mock.patch("{}.{}".format(self.test_class_path, "is_patched"),
                              return_value=True)
        @asynctest.mock.patch("{}.{}".format(self.test_class_path, "a_coroutine"),
                              autospec=True)
        def patched_function(coroutine_mock, is_patched_mock):
            nonlocal called
            called = True

            self.assertIsInstance(Test.is_patched, asynctest.mock.Mock)
            self.assertTrue(Test.is_patched())
            self.assertTrue(asyncio.iscoroutinefunction(coroutine_mock))
            self.assertTrue(asyncio.iscoroutinefunction(Test.a_coroutine))

        patched_function()
        self.assertTrue(called)

    def test_patch_autospec_with_patches_under(self):
        called = False

        @asynctest.mock.patch("{}.{}".format(self.test_class_path, "a_coroutine"),
                              autospec=True)
        @asynctest.mock.patch("{}.{}".format(self.test_class_path, "is_patched"),
                              return_value=True)
        def patched_function(is_patched_mock, coroutine_mock):
            nonlocal called
            called = True

            self.assertIsInstance(Test.is_patched, asynctest.mock.Mock)
            self.assertTrue(Test.is_patched())
            self.assertTrue(asyncio.iscoroutinefunction(coroutine_mock))
            self.assertTrue(asyncio.iscoroutinefunction(Test.a_coroutine))

        patched_function()
        self.assertTrue(called)

    def test_patch_object_autospec(self):
        called = False

        @asynctest.mock.patch.object(Test, "a_coroutine_with_args", autospec=True)
        def patched_function(patched):
            nonlocal called
            called = True

            self.assertTrue(asyncio.iscoroutinefunction(Test.a_coroutine_with_args))
            with self.assertRaisesRegex(TypeError, "arg2"):
                run_coroutine(Test().a_coroutine_with_args("arg"))

            self.assertTrue(run_coroutine(Test().a_coroutine_with_args("arg", "arg2")))

        patched_function()
        self.assertTrue(called)

    def test_patch_multiple_autospec(self):
        called = False
        default = asynctest.mock.DEFAULT

        @asynctest.mock.patch.multiple(Test, autospec=True,
                                       a_coroutine=default,
                                       a_coroutine_with_args=default)
        def patched_function(**patched):
            nonlocal called
            called = True

            with self.assertRaisesRegex(TypeError, "arg2"):
                run_coroutine(Test().a_coroutine_with_args("arg"))

            test = Test()
            self.assertTrue(run_coroutine(test.a_coroutine()))
            self.assertTrue(run_coroutine(test.a_coroutine_with_args("arg", "arg2")))

        patched_function()
        self.assertTrue(called)

#
# patch scopes
#


class patch_scope_TestCase(unittest.TestCase):
    def is_patched(self):
        import test.test_mock
        return test.test_mock.Test().is_patched()

    def second_is_patched(self):
        import test.test_mock
        return test.test_mock.Test().second_is_patched()

    def _test_deactivate_patch_when_generator_init_fails(self, scope):
        @patch_is_patched(scope=scope)
        def a_generator(wrong_number_of_args):
            yield

        try:
            gen = a_generator()
            next(gen)
            self.fail("Exception must raise")
        except TypeError:
            pass

        self.assertFalse(self.is_patched())

    def _test_deactivate_patch_when_generator_exec_fails(self, scope):
        @patch_is_patched(scope=scope)
        @asyncio.coroutine
        def a_coroutine(missing_arg):
            return

        with self.subTest("old style coroutine"):
            @asyncio.coroutine
            def tester():
                try:
                    yield from a_coroutine()
                    self.fail("Exception must raise")
                except TypeError:
                    pass

                self.assertFalse(self.is_patched())

            run_coroutine(tester())

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                lambda missing_arg: None)
            a_new_style_coroutine = patch_is_patched(scope=scope)(
                a_new_style_coroutine)

            @asyncio.coroutine
            def tester():
                try:
                    yield from a_new_style_coroutine()
                    self.fail("Exception must raise")
                except TypeError:
                    pass

                self.assertFalse(self.is_patched())

            run_coroutine(tester())


class patch_dict_scope_TestCase(unittest.TestCase):
    def is_patched(self):
        import test.test_mock
        return test.test_mock.Test().a_dict['is_patched']

    def second_is_patched(self):
        import test.test_mock
        return test.test_mock.Test().a_dict['second_is_patched']

    def second_dict_is_patched(self):
        import test.test_mock
        return test.test_mock.Test().a_second_dict['is_patched']


class Test_patch_dict_decorator_coroutine_or_generator_scope(
        patch_dict_scope_TestCase):
    def test_default_scope_is_global(self):
        @patch_dict_is_patched()
        def a_generator():
            yield self.is_patched()
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertTrue(self.is_patched())
        self.assertTrue(next(gen))

    def test_scope_limited(self):
        @patch_dict_is_patched(scope=asynctest.LIMITED)
        def a_generator():
            yield self.is_patched()
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertFalse(self.is_patched())
        self.assertTrue(next(gen))

    def test_patch_generator_with_multiple_scopes(self):
        with self.subTest("Outher: GLOBAL, inner: LIMITED"):
            @patch_dict_is_patched(scope=asynctest.GLOBAL)
            @patch_dict_second_dict_is_patched(scope=asynctest.LIMITED)
            def a_generator():
                yield (self.is_patched(), self.second_dict_is_patched())
                yield (self.is_patched(), self.second_dict_is_patched())

            gen = a_generator()
            self.assertEqual((True, True), next(gen))
            self.assertEqual(
                (True, False),
                (self.is_patched(), self.second_dict_is_patched()))
            self.assertEqual((True, True), next(gen))

        with self.subTest("Outher: LIMITED, inner: GLOBAL"):
            @patch_dict_is_patched(scope=asynctest.LIMITED)
            @patch_dict_second_dict_is_patched(scope=asynctest.GLOBAL)
            def a_generator():
                yield (self.is_patched(), self.second_dict_is_patched())
                yield (self.is_patched(), self.second_dict_is_patched())

            gen = a_generator()
            self.assertEqual((True, True), next(gen))
            self.assertEqual(
                (False, True),
                (self.is_patched(), self.second_dict_is_patched()))
            self.assertEqual((True, True), next(gen))

    def test_patch_generator_with_multiple_scopes_on_same_dict(self):
        import test.test_mock

        def tester():
            test.test_mock.Test.a_dict['overriden_value'] = True
            for _ in range(2):
                yield (
                    self.is_patched(), self.second_is_patched(),
                    test.test_mock.Test.a_dict.get('overriden_value', False))

        with self.subTest("Outher: GLOBAL, inner: LIMITED"):
            @patch_dict_is_patched(scope=asynctest.GLOBAL)
            @patch_dict_second_is_patched(scope=asynctest.LIMITED)
            def a_generator():
                yield from tester()

            gen = a_generator()
            self.assertEqual((True, True, True), next(gen))
            self.assertEqual((True, False),
                             (self.is_patched(), self.second_is_patched()))
            self.assertNotIn('overriden_value', test.test_mock.Test.a_dict)
            self.assertEqual((True, True, True), next(gen))

        with self.subTest("Outher: LIMITED, inner: GLOBAL"):
            @patch_dict_is_patched(scope=asynctest.LIMITED)
            @patch_dict_second_is_patched(scope=asynctest.GLOBAL)
            def a_generator():
                yield from tester()

            gen = a_generator()
            self.assertEqual((True, True, True), next(gen))
            self.assertEqual((False, True),
                             (self.is_patched(), self.second_is_patched()))
            self.assertNotIn('overriden_value', test.test_mock.Test.a_dict)
            self.assertEqual((True, True, True), next(gen))

    def test_patch_coroutine_with_multiple_scopes(self):
        def tester():
            return (self.is_patched(), self.second_dict_is_patched())

        @asyncio.coroutine
        def tester_couroutine(future):
            before = tester()
            yield from future
            after = tester()
            return before, after

        def run_test(a_coroutine):
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            try:
                future = asyncio.Future(loop=loop)
                task = loop.create_task(a_coroutine(future))
                loop.call_soon(lambda: future.set_result(tester()))
                before, after = loop.run_until_complete(task)
            finally:
                loop.close()

            return before, future.result(), after

        with self.subTest("old style coroutine - Outher: GLOBAL, inner: LIMITED"):
            @patch_dict_is_patched(scope=asynctest.GLOBAL)
            @patch_dict_second_dict_is_patched(scope=asynctest.LIMITED)
            @asyncio.coroutine
            def a_coroutine(future):
                return (yield from tester_couroutine(future))

            before, between, after = run_test(a_coroutine)
            self.assertEqual((True, True), before)
            self.assertEqual((True, False), between)
            self.assertEqual((True, True), after)

        with self.subTest("old style coroutine - Outher: LIMITED, inner: GLOBAL"):
            @patch_dict_is_patched(scope=asynctest.LIMITED)
            @patch_dict_second_dict_is_patched(scope=asynctest.GLOBAL)
            @asyncio.coroutine
            def a_coroutine(future):
                return (yield from tester_couroutine(future))

            before, between, after = run_test(a_coroutine)
            self.assertEqual((True, True), before)
            self.assertEqual((False, True), between)
            self.assertEqual((True, True), after)

        if not _using_await:
            return

        tester_coroutine = _using_await.build_simple_coroutine(tester)

        with self.subTest("new style coroutine - Outher: GLOBAL, inner: LIMITED"):
            a_coroutine = patch_dict_is_patched(scope=asynctest.GLOBAL)(
                patch_dict_second_dict_is_patched(scope=asynctest.LIMITED)(
                    tester_coroutine))

            before, between, after = run_test(a_coroutine)
            self.assertEqual((True, True), before)
            self.assertEqual((True, False), between)
            self.assertEqual((True, True), after)

        with self.subTest("old style coroutine - Outher: LIMITED, inner: GLOBAL"):
            a_coroutine = patch_dict_is_patched(scope=asynctest.LIMITED)(
                patch_dict_second_dict_is_patched(scope=asynctest.GLOBAL)(
                    tester_coroutine))

            before, between, after = run_test(a_coroutine)
            self.assertEqual((True, True), before)
            self.assertEqual((False, True), between)
            self.assertEqual((True, True), after)


class Test_patch_and_patch_dict_scope(unittest.TestCase):
    def test_both_patch_and_patch_dict_with_scope_global(self):
        def test_result():
            import test.test_mock
            instance = test.test_mock.Test()
            return (instance.is_patched(), instance.a_dict['is_patched'])

        with self.subTest("patch and patch.dict"):
            @patch_dict_is_patched(scope=asynctest.GLOBAL)
            @patch_is_patched(scope=asynctest.GLOBAL)
            @asyncio.coroutine
            def a_coroutine():
                return test_result()

            self.assertEqual((True, True), run_coroutine(a_coroutine()))

        with self.subTest("patch.dict and patch"):
            @patch_is_patched(scope=asynctest.GLOBAL)
            @patch_dict_is_patched(scope=asynctest.GLOBAL)
            @asyncio.coroutine
            def a_coroutine():
                return test_result()

            self.assertEqual((True, True), run_coroutine(a_coroutine()))

    def test_both_patch_and_patch_dict_with_scope_limited(self):
        import test.test_mock
        instance = test.test_mock.Test()

        def test_result(instance):
            yield (instance.is_patched(), instance.a_dict['is_patched'])
            yield (instance.is_patched(), instance.a_dict['is_patched'])

        with self.subTest("patch and patch.dict"):
            @patch_dict_is_patched(scope=asynctest.LIMITED)
            @patch_is_patched(scope=asynctest.LIMITED)
            def a_generator(instance):
                yield from test_result(instance)

            gen = a_generator(instance)
            self.assertEqual((True, True), next(gen))
            self.assertEqual((False, False),
                             (instance.is_patched(),
                              instance.a_dict['is_patched']))
            self.assertEqual((True, True), next(gen))

        with self.subTest("patch.dict and patch"):
            @patch_is_patched(scope=asynctest.LIMITED)
            @patch_dict_is_patched(scope=asynctest.LIMITED)
            def a_generator(instance):
                yield from test_result(instance)

            gen = a_generator(instance)
            self.assertEqual((True, True), next(gen))
            self.assertEqual((False, False),
                             (instance.is_patched(),
                              instance.a_dict['is_patched']))
            self.assertEqual((True, True), next(gen))


class Test_patch_decorator_coroutine_or_generator_scope(patch_scope_TestCase):
    # Tests of patch() related to the use of scope=*, with several scopes used
    def test_default_scope_is_global(self):
        @patch_is_patched()
        def a_generator():
            yield self.is_patched()
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertTrue(self.is_patched())
        self.assertTrue(next(gen))

    def test_patch_generator_with_multiple_scopes(self):
        def a_generator():
            yield (self.is_patched(), self.second_is_patched())
            yield (self.is_patched(), self.second_is_patched())

        with self.subTest("Outher: GLOBAL, inner: LIMITED"):
            @patch_is_patched(scope=asynctest.GLOBAL)
            @patch_second_is_patched(scope=asynctest.LIMITED)
            def patched():
                yield from a_generator()

            gen = patched()
            self.assertEqual((True, True), next(gen))
            self.assertTrue(self.is_patched())
            self.assertFalse(self.second_is_patched())
            self.assertEqual((True, True), next(gen))

        with self.subTest("Outher: LIMITED, inner: GLOBAL"):
            @patch_second_is_patched(scope=asynctest.LIMITED)
            @patch_is_patched(scope=asynctest.GLOBAL)
            def patched():
                yield from a_generator()

            gen = patched()
            self.assertEqual((True, True), next(gen))
            self.assertTrue(self.is_patched())
            self.assertFalse(self.second_is_patched())
            self.assertEqual((True, True), next(gen))

    def test_patch_coroutine_with_multiple_scopes(self):
        def set_fut_result(fut):
            fut.set_result((self.is_patched(), self.second_is_patched()))

        @asyncio.coroutine
        def tester(coro_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(set_fut_result, fut)
            before, after = yield from coro_function(fut)
            self.assertEqual((True, True), before)
            self.assertEqual((True, False), fut.result())
            self.assertEqual((True, True), after)
            self.assertFalse(self.is_patched())
            self.assertFalse(self.second_is_patched())

        with self.subTest("old style coroutine - Outher: GLOBAL, inner: LIMITED"):
            @patch_is_patched(scope=asynctest.GLOBAL)
            @patch_second_is_patched(scope=asynctest.LIMITED)
            def a_coroutine(fut):
                before = (self.is_patched(), self.second_is_patched())
                yield from fut
                after = (self.is_patched(), self.second_is_patched())
                return before, after

            run_coroutine(tester(a_coroutine))

        with self.subTest("old style coroutine - Outher: LIMITED, inner: GLOBAL"):
            @patch_second_is_patched(scope=asynctest.LIMITED)
            @patch_is_patched(scope=asynctest.GLOBAL)
            def a_coroutine(fut):
                before = (self.is_patched(), self.second_is_patched())
                yield from fut
                after = (self.is_patched(), self.second_is_patched())
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine - Outher: GLOBAL, inner: LIMITED"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                lambda: (self.is_patched(), self.second_is_patched()))
            a_new_style_coroutine = patch_second_is_patched(scope=asynctest.LIMITED)(
                a_new_style_coroutine)
            a_new_style_coroutine = patch_is_patched(scope=asynctest.GLOBAL)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))

        with self.subTest("new style coroutine - Outher: LIMITED, inner: GLOBAL"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                lambda: (self.is_patched(), self.second_is_patched()))
            a_new_style_coroutine = patch_is_patched(scope=asynctest.GLOBAL)(
                a_new_style_coroutine)
            a_new_style_coroutine = patch_second_is_patched(scope=asynctest.LIMITED)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))


class Test_patch_decorator_coroutine_or_generator_scope_GLOBAL(patch_scope_TestCase):
    # Tests of patch() using scope=GLOBAL
    def test_deactivate_patch_when_generator_init_fails(self):
        self._test_deactivate_patch_when_generator_init_fails(asynctest.GLOBAL)

    def test_deactivate_patch_when_generator_exec_fails(self):
        self._test_deactivate_patch_when_generator_exec_fails(asynctest.GLOBAL)

    def test_patch_generator_during_its_lifetime(self):
        @patch_is_patched(scope=asynctest.GLOBAL)
        def a_generator():
            yield self.is_patched()
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertTrue(self.is_patched())
        self.assertTrue(next(gen))
        # exhaust the generator
        try:
            next(gen)
            self.fail("Coroutine must be stopped")
        except StopIteration:
            pass
        self.assertFalse(self.is_patched())

    def test_patch_coroutine_during_its_lifetime(self):
        def set_fut_result(fut):
            fut.set_result(self.is_patched())

        @asyncio.coroutine
        def tester(coro_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(set_fut_result, fut)
            before, after = yield from coro_function(fut)
            self.assertTrue(before)
            self.assertTrue(fut.result())
            self.assertTrue(after)
            self.assertFalse(self.is_patched())

        with self.subTest("old style coroutine"):
            @patch_is_patched(scope=asynctest.GLOBAL)
            def a_coroutine(fut):
                before = self.is_patched()
                yield from fut
                after = self.is_patched()
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                self.is_patched)
            a_new_style_coroutine = patch_is_patched(scope=asynctest.GLOBAL)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))

    # It's really hard to test this behavior for a coroutine, but I assume it's
    # fine as long as the implementation is shared with a generator. Also, it's
    # really hard to fall in a case like this one with a coroutine.
    def test_patch_stopped_when_generator_is_collected(self):
        @patch_is_patched(scope=asynctest.GLOBAL)
        def a_generator():
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertTrue(self.is_patched())
        del gen
        self.assertFalse(self.is_patched())

    def test_multiple_patches_on_generator(self):
        @patch_second_is_patched(scope=asynctest.GLOBAL)
        @patch_is_patched(scope=asynctest.GLOBAL)
        def a_generator():
            yield self.is_patched() and self.second_is_patched()
            yield self.is_patched() and self.second_is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertTrue(self.is_patched())
        self.assertTrue(self.second_is_patched())
        self.assertTrue(next(gen))
        # exhaust the generator
        try:
            next(gen)
            self.fail("Coroutine must be stopped")
        except StopIteration:
            pass
        self.assertFalse(self.is_patched())
        self.assertFalse(self.second_is_patched())

    def test_multiple_patches_on_coroutine(self):
        def set_fut_result(fut):
            fut.set_result((self.is_patched(), self.second_is_patched()))

        @asyncio.coroutine
        def tester(coro_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(set_fut_result, fut)
            before, after = yield from coro_function(fut)
            self.assertEqual((True, True), before)
            self.assertEqual((True, True), fut.result())
            self.assertEqual((True, True), after)
            self.assertFalse(self.is_patched())
            self.assertFalse(self.second_is_patched())

        with self.subTest("old style coroutine"):
            @patch_second_is_patched(scope=asynctest.GLOBAL)
            @patch_is_patched(scope=asynctest.GLOBAL)
            def a_coroutine(fut):
                before = (self.is_patched(), self.second_is_patched())
                yield from fut
                after = (self.is_patched(), self.second_is_patched())
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                lambda: (self.is_patched(), self.second_is_patched()))
            a_new_style_coroutine = patch_second_is_patched(scope=asynctest.GLOBAL)(
                a_new_style_coroutine)
            a_new_style_coroutine = patch_is_patched(scope=asynctest.GLOBAL)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))


class Test_patch_decorator_coroutine_or_generator_scope_LIMITED(patch_scope_TestCase):
    # Tests of patch() using scope=LIMITED
    def test_deactivate_patch_when_generator_init_fails(self):
        self._test_deactivate_patch_when_generator_init_fails(asynctest.LIMITED)

    def test_deactivate_patch_when_generator_exec_fails(self):
        self._test_deactivate_patch_when_generator_exec_fails(asynctest.LIMITED)

    def test_patch_generator_only_when_running(self):
        @patch_is_patched(scope=asynctest.LIMITED)
        def a_generator():
            yield self.is_patched()
            yield self.is_patched()

        gen = a_generator()
        self.assertTrue(next(gen))
        self.assertFalse(self.is_patched())
        self.assertTrue(next(gen))

    def test_patch_coroutine_only_when_running(self):
        def set_fut_result(fut):
            fut.set_result(self.is_patched())

        @asyncio.coroutine
        def tester(coro_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(set_fut_result, fut)
            before, after = yield from coro_function(fut)
            self.assertTrue(before)
            self.assertFalse(fut.result())
            self.assertTrue(after)

        with self.subTest("old style coroutine"):
            @patch_is_patched(scope=asynctest.LIMITED)
            def a_coroutine(fut):
                before = self.is_patched()
                yield from fut
                after = self.is_patched()
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                self.is_patched)
            a_new_style_coroutine = patch_is_patched(scope=asynctest.LIMITED)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))

    def test_patched_coroutine_with_mock_args(self):
        @asynctest.mock.patch('test.test_mock.Test', side_effect=lambda: None,
                              scope=asynctest.LIMITED)
        @asyncio.coroutine
        def a_coroutine(mock):
            loop = asyncio.get_event_loop()
            self.assertIs(mock, Test)
            yield from asyncio.sleep(0, loop=loop)
            self.assertIs(mock, Test)
            yield from asyncio.sleep(0, loop=loop)
            self.assertIs(mock, Test)

        run_coroutine(a_coroutine())

    def test_multiple_patches_on_coroutine(self):
        def set_fut_result(fut):
            fut.set_result((self.is_patched(), self.second_is_patched()))

        @asyncio.coroutine
        def tester(coro_function):
            loop = asyncio.get_event_loop()
            fut = asyncio.Future(loop=loop)
            loop.call_soon(set_fut_result, fut)
            before, after = yield from coro_function(fut)
            self.assertEqual((True, True), before)
            self.assertEqual((False, False), fut.result())
            self.assertEqual((True, True), after)
            self.assertFalse(self.is_patched())
            self.assertFalse(self.second_is_patched())

        with self.subTest("old style coroutine"):
            @patch_second_is_patched(scope=asynctest.LIMITED)
            @patch_is_patched(scope=asynctest.LIMITED)
            def a_coroutine(fut):
                before = (self.is_patched(), self.second_is_patched())
                yield from fut
                after = (self.is_patched(), self.second_is_patched())
                return before, after

            run_coroutine(tester(a_coroutine))

        if not _using_await:
            return

        with self.subTest("new style coroutine"):
            a_new_style_coroutine = _using_await.build_simple_coroutine(
                lambda: (self.is_patched(), self.second_is_patched()))
            a_new_style_coroutine = patch_second_is_patched(scope=asynctest.LIMITED)(
                a_new_style_coroutine)
            a_new_style_coroutine = patch_is_patched(scope=asynctest.LIMITED)(
                a_new_style_coroutine)
            run_coroutine(tester(a_new_style_coroutine))


class Test_return_once(unittest.TestCase):
    def test_default_value(self):
        iterator = asynctest.mock.return_once("ProbeValue")
        self.assertEqual("ProbeValue", next(iterator))
        for _ in range(3):
            self.assertIsNone(next(iterator))

    def test_then(self):
        iterator = asynctest.mock.return_once("ProbeValue", "ThenValue")
        self.assertEqual("ProbeValue", next(iterator))
        for _ in range(2):
            self.assertEqual("ThenValue", next(iterator))

        iterator = asynctest.mock.return_once("ProbeValue", then="ThenValue")
        self.assertEqual("ProbeValue", next(iterator))
        self.assertEqual("ThenValue", next(iterator))

    def test_with_side_effect_default(self):
        mock = asynctest.Mock(side_effect=asynctest.mock.return_once("ProbeValue"))
        self.assertEqual("ProbeValue", mock())
        for _ in range(3):
            self.assertIsNone(mock())

    def test_with_side_effect_then(self):
        side_effect = asynctest.mock.return_once("ProbeValue", "ThenValue")
        mock = asynctest.Mock(side_effect=side_effect)
        self.assertEqual("ProbeValue", mock())
        for _ in range(2):
            self.assertEqual("ThenValue", mock())

    def test_with_side_effect_raises(self):
        mock = asynctest.mock.Mock(side_effect=asynctest.mock.return_once(Exception))
        self.assertRaises(Exception, mock)
        self.assertIsNone(mock())

    def test_with_side_effect_raises_then(self):
        side_effect = asynctest.mock.return_once("ProbeValue", BlockingIOError)
        mock = asynctest.mock.Mock(side_effect=side_effect)
        self.assertEqual("ProbeValue", mock())
        for _ in range(2):
            self.assertRaises(BlockingIOError, mock)

    def test_with_side_effect_raises_all(self):
        side_effect = asynctest.mock.return_once(Exception, BlockingIOError)
        mock = asynctest.mock.Mock(side_effect=side_effect)
        self.assertRaises(Exception, mock)
        for _ in range(2):
            self.assertRaises(BlockingIOError, mock)


class Test_create_autospec(unittest.TestCase):
    def test_generator_and_coroutine_is_instance_of_FunctionType(self):
        # this test is somewhat a forward compatibility test: if ever
        # unittest.mock.FunctionTypes doesn't detect generators and coroutines
        # as instance of these types, we need to fix it in asynctest.
        def gen():
            yield from range(10)

        @asyncio.coroutine
        def coroutine():
            yield from asyncio.sleep(0)

        self.assertIsInstance(gen, unittest.mock.FunctionTypes)
        self.assertIsInstance(coroutine, unittest.mock.FunctionTypes)
        if _using_await:
            self.assertIsInstance(_using_await.transform(coroutine),
                                  unittest.mock.FunctionTypes)

    # create_autospec:
    # * Ensure we check the signature of the coroutine function (and/or
    #   generator)
    # * Ensure a coroutine function attribute of a mock create with
    #   create_autospec is a CoroutineMock
    # * Ensure an instance of a class mocked from create_autospec will be an
    #   asynctest mock, and its coroutine attribute will be mocked by
    #   a CoroutineMock
    # * Ensure all expected create_autospec tests still run fine
    #
    # Also test cases where create_autospec is used (_patch, etc)
    def test_autospec_returns_asynctest_mocks(self):
        def a_generator():
            yield from range(10)

        cases = {
            "non callable": "Constant value",
            "callable class": Test,
            "non callable instance": Test(),
            "callable function": Test.a_function,
            "callable generator": a_generator,
        }

        for name, value in cases.items():
            with self.subTest(name):
                mock = asynctest.mock.create_autospec(value)
                unittest_mock = unittest.mock.create_autospec(value)
                try:
                    expected_type = getattr(asynctest.mock,
                                            type(unittest_mock).__name__)
                except AttributeError:
                    # The type of returned object is not readable as a mock
                    # This happens with mocks updated by _set_signature
                    expected_type = type(unittest_mock)
                self.assertIsInstance(mock, expected_type)
                self.assertNotIsInstance(mock, asynctest.mock.CoroutineMock)

    def test_autospec_of_coroutine_function_is_coroutinefunction(self):
        mock = asynctest.mock.create_autospec(Test.a_function)
        self.assertFalse(asyncio.iscoroutinefunction(mock))

        mock = asynctest.mock.create_autospec(Test.a_coroutine)
        self.assertTrue(asyncio.iscoroutinefunction(mock))

        if _using_await:
            mock = asynctest.mock.create_autospec(Test.an_async_coroutine)
            self.assertTrue(asyncio.iscoroutinefunction(mock))

    def test_autospec_attributes_being_coroutine_functions(self):
        mock = asynctest.mock.create_autospec(Test)
        self.assertFalse(asyncio.iscoroutinefunction(mock))
        self.assertFalse(asyncio.iscoroutinefunction(mock.a_function))
        self.assertTrue(asyncio.iscoroutinefunction(mock.a_coroutine))
        if _using_await:
            self.assertTrue(asyncio.iscoroutinefunction(mock.an_async_coroutine))

    def test_create_autospec_on_coroutine_with_return_value(self):
        mock = asynctest.mock.create_autospec(Test.a_coroutine,
                                              return_value="PROBE")
        self.assertEqual("PROBE", run_coroutine(mock(None)))

        if _using_await:
            mock = asynctest.mock.create_autospec(Test.an_async_coroutine,
                                                  return_value="PROBE")
            self.assertEqual("PROBE", run_coroutine(mock(None)))

    def test_create_autospec_on_coroutine_with_iterable_side_effect(self):
        coroutines = [Test.a_coroutine]
        if _using_await:
            coroutines.append(Test.an_async_coroutine)

        for a_coroutine in coroutines:
            mock = asynctest.mock.create_autospec(
                a_coroutine, side_effect=("PROBE1", "PROBE2"))
            self.assertEqual("PROBE1", run_coroutine(mock(None)))
            self.assertEqual("PROBE2", run_coroutine(mock(None)))

    def test_create_autospec_on_coroutine_with_exception_side_effect(self):
        coroutines = [Test.a_coroutine]
        if _using_await:
            coroutines.append(Test.an_async_coroutine)

        for a_coroutine in coroutines:
            mock = asynctest.mock.create_autospec(a_coroutine,
                                                  side_effect=ProbeException)
            with self.assertRaises(ProbeException):
                run_coroutine(mock(None))

    def test_create_autospec_on_coroutine_with_coroutine_side_effect(self):
        coroutines = [Test.a_coroutine]
        if _using_await:
            coroutines.append(Test.an_async_coroutine)

        for a_coroutine in coroutines:
            mock = asynctest.mock.create_autospec(
                a_coroutine, side_effect=asyncio.coroutine(lambda r: r))
            self.assertEqual("PROBE", run_coroutine(mock("PROBE")))

    def test_create_autospec_on_coroutine_with_instance_raises_RuntimeError(self):
        with self.assertRaises(RuntimeError):
            asynctest.mock.create_autospec(Test.a_coroutine, instance=True)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
Helpers
-------

Helper functions and coroutines for :mod:`asynctest`.
"""

import asyncio


@asyncio.coroutine
def exhaust_callbacks(loop):
    """
    Run the loop until all ready callbacks are executed.

    The coroutine doesn't wait for callbacks scheduled in the future with
    :meth:`~asyncio.BaseEventLoop.call_at()` or
    :meth:`~asyncio.BaseEventLoop.call_later()`.

    :param loop: event loop
    """
    while loop._ready:
        yield from asyncio.sleep(0, loop=loop)



# coding: utf-8
"""
asynctest setup script.

We rely on setuptools >= 30.3.0 to read setup.cfg, but provide a minimal
support with distutils.
"""

args = {
    "name": "asynctest",
    "packages": ["asynctest"],
}


def read_version():
    import configparser
    import os

    with open(os.path.join(os.path.dirname(__file__), "setup.cfg")) as cfg:
        config = configparser.ConfigParser()
        config.read_file(cfg, source="setup.cfg")
        return config['metadata']['version']


try:
    # We don't use this method, but it allows to detect if setuptools will read
    # the setup.cfg file or if we need to find the version by ourselves.
    from setuptools.config import read_configuration  # noqa
except ImportError:
        args['version'] = read_version()


if __name__ == "__main__":
    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup

    setup(**args)



# coding: utf-8

import asyncio
import copy
import selectors
import functools
import os
import socket
try:
    import ssl
except ImportError:
    ssl = None
import sys
import unittest

import asynctest


class Selector_TestCase(unittest.TestCase):
    def setUp(self):
        asynctest.selector.FileDescriptor.next_fd = 0


class Test_FileDescriptor(Selector_TestCase):
    def test_is_an_int(self):
        self.assertIsInstance(asynctest.selector.FileDescriptor(), int)

    def test_init_increments_value(self):
        self.assertEqual(0, asynctest.selector.FileDescriptor())
        self.assertEqual(1, asynctest.selector.FileDescriptor())

        self.assertNotEqual(asynctest.selector.FileDescriptor(),
                            asynctest.selector.FileDescriptor())

    def test_init_increments_value_with_fixed_value(self):
        self.assertEqual(5, asynctest.selector.FileDescriptor(5))
        self.assertEqual(6, asynctest.selector.FileDescriptor())


class Test_FileMock(Selector_TestCase):
    def test_fileno_returns_FileDescriptor(self):
        self.assertIsInstance(asynctest.selector.FileMock().fileno(),
                              asynctest.selector.FileDescriptor)


class Test_SocketMock(Selector_TestCase):
    def test_is_socket(self):
        self.assertIsInstance(asynctest.selector.SocketMock(), socket.socket)


if ssl:
    class Test_SSLSocketMock(Selector_TestCase):
        def test_is_ssl_socket(self):
            self.assertIsInstance(asynctest.selector.SSLSocketMock(),
                                  ssl.SSLSocket)


def selector_subtest(method):
    @functools.wraps(method)
    def wrapper(self):
        with self.subTest(test='without_selector'):
            method(self, asynctest.selector.TestSelector(), None)

        with self.subTest(test='with_selector'):
            mock = unittest.mock.Mock(selectors.BaseSelector)
            method(self, asynctest.selector.TestSelector(mock), mock)

    return wrapper


class Test_TestSelector(Selector_TestCase):
    @selector_subtest
    def test_register_mock(self, selector, selector_mock):
        mock = asynctest.selector.FileMock()
        key = selector.register(mock, selectors.EVENT_READ, "data")

        self.assertEqual(key, selector.get_map()[mock])

        if selector_mock:
            self.assertFalse(selector_mock.register.called)

    @selector_subtest
    def test_register_fileno(self, selector, selector_mock):
        with open(os.devnull, 'r') as devnull:
            if selector_mock:
                selector_mock.register.return_value = selectors.SelectorKey(
                    devnull, devnull.fileno(), selectors.EVENT_READ, "data"
                )

            key = selector.register(devnull, selectors.EVENT_READ, "data")

            self.assertEqual(key, selector.get_map()[devnull])

            if selector_mock:
                selector_mock.register.assert_called_with(devnull,
                                                          selectors.EVENT_READ,
                                                          "data")

    @selector_subtest
    def test_unregister_mock(self, selector, selector_mock):
        mock = asynctest.selector.FileMock()
        selector.register(mock, selectors.EVENT_READ, "data")

        selector.unregister(mock)

        self.assertNotIn(mock, selector.get_map())
        self.assertNotIn(mock.fileno(), selector.get_map())

        if selector_mock:
            self.assertFalse(selector_mock.unregister.called)

    @selector_subtest
    def test_unregister_fileno(self, selector, selector_mock):
        with open(os.devnull, 'r') as devnull:
            if selector_mock:
                key = selectors.SelectorKey(devnull, devnull.fileno(),
                                            selectors.EVENT_READ, "data")
                selector_mock.register.return_value = key
                selector_mock.unregister.return_value = key

            selector.register(devnull, selectors.EVENT_READ, "data")

            selector.unregister(devnull)

            self.assertNotIn(devnull, selector.get_map())
            self.assertNotIn(devnull.fileno(), selector.get_map())

    @selector_subtest
    def test_modify_mock(self, selector, selector_mock):
        mock = asynctest.selector.FileMock()

        original_key = selector.register(mock, selectors.EVENT_READ, "data")
        # modify may update the original key, keep a copy
        original_key = copy.copy(original_key)

        RW = selectors.EVENT_READ | selectors.EVENT_WRITE

        key = selector.modify(mock, RW, "data")

        self.assertNotEqual(original_key, key)
        self.assertEqual(key, selector.get_map()[mock])

    @selector_subtest
    def test_modify_fileno(self, selector, selector_mock):
        with open(os.devnull, 'r') as devnull:
            if selector_mock:
                selector_mock.modify.return_value = selectors.SelectorKey(
                    devnull, devnull.fileno(), selectors.EVENT_READ, "data2"
                )

            original_key = selector.register(devnull, selectors.EVENT_READ, "data")
            # modify may update the original key, keep a copy
            original_key = copy.copy(original_key)

            key = selector.modify(devnull, selectors.EVENT_READ, "data2")

            self.assertNotEqual(original_key, key)
            self.assertEqual(key, selector.get_map()[devnull])

            if selector_mock:
                selector_mock.modify.assert_called_with(devnull, selectors.EVENT_READ, "data2")

    @selector_subtest
    def test_modify_fd(self, selector, selector_mock):
        fd = 1

        if selector_mock:
            selector_mock.modify.return_value = selectors.SelectorKey(
                fd, fd, selectors.EVENT_READ, "data2"
            )

        original_key = selector.register(fd, selectors.EVENT_READ, "data")
        original_key = copy.copy(original_key)

        key = selector.modify(fd, selectors.EVENT_READ, "data2")

        self.assertNotEqual(original_key, key)
        self.assertEqual(key, selector.get_map()[fd])

        if selector_mock:
            selector_mock.modify.assert_called_with(fd, selectors.EVENT_READ, "data2")

    @selector_subtest
    def test_modify_but_selector_raises(self, selector, selector_mock):
        if not selector_mock:
            return

        exception = RuntimeError()
        selector_mock.modify.side_effect = exception

        with open(os.devnull, 'r') as devnull:
            selector.register(devnull, selectors.EVENT_READ, "data")

            with self.assertRaises(type(exception)) as ctx:
                selector.modify(devnull, selectors.EVENT_READ, "data2")

            self.assertIs(exception, ctx.exception)
            self.assertNotIn(devnull, selector.get_map())

    @selector_subtest
    def test_select(self, selector, selector_mock):
        if selector_mock:
            selector_mock.select.return_value = ["ProbeValue"]
            self.assertEqual(["ProbeValue"], selector.select(5))
            selector_mock.select.assert_called_with(5)
        else:
            self.assertEqual([], selector.select())

    @selector_subtest
    def test_close(self, selector, selector_mock):
        if not selector_mock:
            return

        selector.close()
        selector_mock.close.assert_called_with()


class Test_set_read_write_ready(Selector_TestCase):
    def setUp(self):
        super().setUp()

        self.loop = asyncio.new_event_loop()
        self.loop._selector = asynctest.selector.TestSelector(self.loop._selector)
        self.addCleanup(self.loop.close)
        self.mock = asynctest.selector.FileMock()

        # Older versions of asyncio may complain with PYTHONASYNCIODEBUG=1
        if sys.version_info < (3, 5):
            asyncio.set_event_loop(self.loop)

    def test_nothing_scheduled(self):
        # nothing will happen (no exception)
        for mode in ('read', 'write'):
            with self.subTest(mode=mode):
                getattr(asynctest.selector, 'set_{}_ready'.format(mode))(self.mock, self.loop)
                self.loop.run_until_complete(asyncio.sleep(0, loop=self.loop))

    def test_callback_scheduled(self):
        for mode in ('read', 'write'):
            with self.subTest(mode=mode):
                future = asyncio.Future(loop=self.loop)
                callback_mock = unittest.mock.Mock()

                # We need at least two iterations of the loop
                self.loop.call_soon(self.loop.call_soon, future.set_result, None)

                getattr(self.loop, 'add_{}er'.format(mode.strip('e')))(self.mock, callback_mock)
                getattr(asynctest.selector, 'set_{}_ready'.format(mode))(self.mock, self.loop)

                self.loop.run_until_complete(future)
                callback_mock.assert_called_with()

    def test_callback_scheduled_during_current_iteration(self):
        for mode in ('read', 'write'):
            with self.subTest(mode=mode):
                future = asyncio.Future(loop=self.loop)
                callback_mock = unittest.mock.Mock()

                # We need at least two iterations of the loop
                self.loop.call_soon(self.loop.call_soon, future.set_result, None)

                self.loop.call_soon(getattr(self.loop, 'add_{}er'.format(mode.strip('e'))),
                                    self.mock, callback_mock)
                getattr(asynctest.selector, 'set_{}_ready'.format(mode))(self.mock, self.loop)

                self.loop.run_until_complete(future)
                callback_mock.assert_called_with()


@unittest.mock.patch.dict('asynctest._fail_on.DEFAULTS', clear=True,
                          active_selector_callbacks=True)
class Test_fail_on_active_selector_callbacks(Selector_TestCase):
    def test_passes_without_callbacks_set(self):
        class TestCase(asynctest.TestCase):
            def runTest(self):
                pass

        TestCase().debug()

    def test_passes_when_no_callbacks_left(self):
        class TestCase(asynctest.TestCase):
            def runTest(self):
                mock = asynctest.selector.FileMock()
                self.loop.add_reader(mock, lambda: None)
                self.loop.remove_reader(mock)

        TestCase().debug()

    def test_events_watched_outside_test_are_ignored(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            mock = asynctest.selector.FileMock()
            loop.add_reader(mock, lambda: None)
            self.addCleanup(loop.remove_reader, mock)

            class TestCase(asynctest.TestCase):
                use_default_loop = False

                def runTest(self):
                    pass

            TestCase().debug()
        finally:
            loop.close()
            asyncio.set_event_loop(None)

    def test_fail_on_active_selector_callbacks_on_mock_files(self):
        class TestCase(asynctest.TestCase):
            def runTest(self):
                mock = asynctest.selector.FileMock()
                self.loop.add_reader(mock, lambda: None)
                # it's too late to check that during cleanup
                self.addCleanup(self.loop.remove_reader, mock)

        with self.assertRaisesRegex(AssertionError, "some events watched "
                                    "during the tests were not removed"):
            TestCase().debug()

    def test_fail_on_original_selector_callback(self):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        try:
            with unittest.mock.patch.object(loop, "_selector") as mock:
                class TestCase(asynctest.TestCase):
                    use_default_loop = True

                    def runTest(self):
                        # add a dummy event
                        handle = asyncio.Handle(lambda: None, (), self.loop)
                        key = selectors.SelectorKey(1, 1, selectors.EVENT_READ,
                                                    (handle, None))
                        mock.get_map.return_value = {1: key}

                with self.assertRaisesRegex(AssertionError,
                                            "some events watched during the "
                                            "tests were not removed"):
                    TestCase().debug()
        finally:
            loop.close()
            asyncio.set_event_loop(None)



# coding: utf-8

import asyncio
import unittest

import asynctest


class TestExhaust(asynctest.TestCase):
    @asyncio.coroutine
    def wait_for(self, coro):
        return (yield from asyncio.wait_for(coro, loop=self.loop, timeout=1))

    def test_exhaust_callbacks_nothing_to_wait(self):
        # Nothing ready, do nothing (must not timeout)
        yield from self.wait_for(asynctest.helpers.exhaust_callbacks(self.loop))

    def test_exhaust_callbacks_one_callback(self):
        fut = asyncio.Future(loop=self.loop)
        self.loop.call_soon(fut.set_result, None)

        # A callback has been scheduled
        yield from self.wait_for(asynctest.helpers.exhaust_callbacks(self.loop))
        self.assertTrue(fut.done())

    def test_exhaust_callbacks_cascading_callbacks(self):
        # A callback has been scheduled, then another (while the 1st is
        # running), we must wait for both
        fut = asyncio.Future(loop=self.loop)
        fut2 = asyncio.Future(loop=self.loop)

        fut.add_done_callback(lambda _: fut2.set_result(None))
        self.loop.call_soon(fut.set_result, None)
        yield from self.wait_for(asynctest.helpers.exhaust_callbacks(self.loop))
        self.assertTrue(fut.done())
        self.assertTrue(fut2.done())


if __name__ == "__main__":
    unittest.main()
